# -*- coding: utf-8 -*-

import unittest
from unittest.mock import MagicMock, patch

from quasarr.downloads import submit_final_download_urls
from quasarr.downloads.mirror_filters import (
    filter_final_download_urls,
    normalize_mirror_token,
)


class NormalizeMirrorTokenTests(unittest.TestCase):
    def test_normalizes_whitelist_names(self):
        self.assertEqual(normalize_mirror_token("DDownload"), "ddownload")
        self.assertEqual(normalize_mirror_token("Keep2Share"), "keep2share")

    def test_normalizes_alias_domains_and_subdomains(self):
        cases = {
            "https://api.ddl.to/container": "ddownload",
            "https://s42.rg.to/file/abc": "rapidgator",
            "https://subdomain.nitroflare.com/view/test": "nitroflare",
            "https://download.ifolder.com.ua/file/123": "turbobit",
            "https://mega.co.nz/file/test": "mega",
            "https://clickndownload.space/abcdef": "clicknupload",
        }

        for raw_value, expected in cases.items():
            with self.subTest(raw_value=raw_value):
                self.assertEqual(normalize_mirror_token(raw_value), expected)


class FilterFinalDownloadUrlsTests(unittest.TestCase):
    def test_keeps_only_allowed_final_urls(self):
        result = filter_final_download_urls(
            [
                "https://rapidgator.net/file/abc",
                "https://cdn.ddownload.com/xyz",
                "https://nitroflare.com/view/test",
            ],
            ["DDownload"],
        )

        self.assertEqual(result["urls"], ["https://cdn.ddownload.com/xyz"])
        self.assertEqual(
            {item["token"] for item in result["dropped"]},
            {"rapidgator", "nitroflare"},
        )


class SubmitFinalDownloadUrlsTests(unittest.TestCase):
    @patch("quasarr.downloads.download_package", return_value=True)
    @patch(
        "quasarr.downloads.get_download_category_mirrors", return_value=["DDownload"]
    )
    @patch("quasarr.downloads.get_download_category_from_package_id", return_value="tv")
    def test_submit_uses_filtered_urls(
        self,
        mock_get_category,
        mock_get_mirrors,
        mock_download_package,
    ):
        shared_state = MagicMock()

        result = submit_final_download_urls(
            shared_state,
            [
                "https://rapidgator.net/file/abc",
                "https://mirror.ddownload.com/file/def",
            ],
            "Example.Release",
            "",
            "Quasarr_tv_deadbeefdeadbeefdeadbeefdeadbeef",
        )

        self.assertTrue(result["success"])
        self.assertEqual(result["links"], ["https://mirror.ddownload.com/file/def"])
        mock_download_package.assert_called_once_with(
            ["https://mirror.ddownload.com/file/def"],
            "Example.Release",
            "",
            "Quasarr_tv_deadbeefdeadbeefdeadbeefdeadbeef",
            shared_state,
        )

    @patch("quasarr.downloads.download_package")
    @patch("quasarr.downloads.fail", return_value={"success": True, "failed": True})
    @patch(
        "quasarr.downloads.get_download_category_mirrors", return_value=["DDownload"]
    )
    @patch("quasarr.downloads.get_download_category_from_package_id", return_value="tv")
    def test_submit_persists_failure_when_no_allowed_links_remain(
        self,
        mock_get_category,
        mock_get_mirrors,
        mock_fail,
        mock_download_package,
    ):
        protected_db = MagicMock()
        shared_state = MagicMock()
        shared_state.get_db.return_value = protected_db

        result = submit_final_download_urls(
            shared_state,
            ["https://rapidgator.net/file/abc"],
            "Example.Release",
            "",
            "Quasarr_tv_deadbeefdeadbeefdeadbeefdeadbeef",
            remove_protected=True,
        )

        self.assertFalse(result["success"])
        self.assertTrue(result["persisted_failure"])
        protected_db.delete.assert_called_once_with(
            "Quasarr_tv_deadbeefdeadbeefdeadbeefdeadbeef"
        )
        mock_fail.assert_called_once()
        mock_download_package.assert_not_called()

    @patch("quasarr.downloads.download_package", return_value=True)
    @patch(
        "quasarr.downloads.get_download_category_mirrors", return_value=["DDownload"]
    )
    @patch("quasarr.downloads.get_download_category_from_package_id", return_value="tv")
    def test_submit_removes_protected_package_after_success_when_requested(
        self,
        mock_get_category,
        mock_get_mirrors,
        mock_download_package,
    ):
        protected_db = MagicMock()
        shared_state = MagicMock()
        shared_state.get_db.return_value = protected_db

        result = submit_final_download_urls(
            shared_state,
            ["https://mirror.ddownload.com/file/def"],
            "Example.Release",
            "",
            "Quasarr_tv_deadbeefdeadbeefdeadbeefdeadbeef",
            remove_protected=True,
        )

        self.assertTrue(result["success"])
        protected_db.delete.assert_called_once_with(
            "Quasarr_tv_deadbeefdeadbeefdeadbeefdeadbeef"
        )
        mock_download_package.assert_called_once()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re
from dataclasses import dataclass
from typing import List, Optional

from quasarr.constants import LANGUAGE_TO_ALPHA2, SUBTITLE_TOKEN_BY_ALPHA2
from quasarr.providers.utils import sanitize_title


@dataclass
class ReleaseInfo:
    release_title: Optional[str]
    audio_langs: List[str]
    subtitle_langs: List[str]
    episode_title: Optional[str]
    resolution: str
    audio: str
    video: str
    source: str
    release_group: str
    season_part: Optional[int]
    season: Optional[int]
    episode_min: Optional[int]
    episode_max: Optional[int]


def subtitle_lang_to_alpha2(lang: str) -> Optional[str]:
    if not lang:
        return None

    normalized = re.sub(r"[^a-z]", "", lang.lower())
    if not normalized:
        return None

    mapped = LANGUAGE_TO_ALPHA2.get(normalized)
    if mapped:
        return mapped

    if len(normalized) == 2:
        return normalized.upper()

    return None


def subtitle_tokens(subtitle_langs: List[str]) -> List[str]:
    tokens: List[str] = []
    for lang in subtitle_langs:
        code = subtitle_lang_to_alpha2(lang)
        token = SUBTITLE_TOKEN_BY_ALPHA2.get(code) if code else None
        if token and token not in tokens:
            tokens.append(token)
    return tokens


def inject_subtitle_tokens_in_title(title: str, subtitle_langs: List[str]) -> str:
    """
    Canonical subtitle marker format:
    - `GerSub`, `EngSub`, `JapSub`
    """
    has_group = re.search(r"-[^.\s]+$", title)
    title_without_group = title
    group_suffix = ""
    if has_group:
        title_without_group, _, group = title.rpartition("-")
        group_suffix = f"-{group}"

    tokens = [t for t in title_without_group.split(".") if t]

    subtitle_token_by_lower = {
        token.lower(): token for token in SUBTITLE_TOKEN_BY_ALPHA2.values()
    }
    existing_subtitle_tokens: List[str] = []
    first_existing_marker_idx = None

    def add_unique_subtitle_token(token: str, target: List[str]) -> None:
        canonical = subtitle_token_by_lower.get(token.lower())
        if canonical and canonical not in target:
            target.append(canonical)

    cleaned_tokens: List[str] = []
    for token in tokens:
        if token.lower() in subtitle_token_by_lower:
            if first_existing_marker_idx is None:
                first_existing_marker_idx = len(cleaned_tokens)
            add_unique_subtitle_token(token, existing_subtitle_tokens)
            continue
        cleaned_tokens.append(token)
    tokens = cleaned_tokens

    first_subbed_idx = None
    cleaned_tokens = []
    for token in tokens:
        if token.lower() == "subbed":
            if first_subbed_idx is None:
                first_subbed_idx = len(cleaned_tokens)
            continue
        cleaned_tokens.append(token)
    tokens = cleaned_tokens

    resolved_subtitle_tokens = subtitle_tokens(subtitle_langs)
    if not resolved_subtitle_tokens:
        resolved_subtitle_tokens = existing_subtitle_tokens

    if resolved_subtitle_tokens:
        if first_subbed_idx is not None:
            insert_at = first_subbed_idx
        elif first_existing_marker_idx is not None:
            insert_at = first_existing_marker_idx
        else:
            insert_at = len(tokens)
        insert_at = max(0, min(insert_at, len(tokens)))
        tokens[insert_at:insert_at] = resolved_subtitle_tokens

    normalized_title = ".".join(tokens) + group_suffix
    return sanitize_title(normalized_title)


def guess_release_title(page_title: str, release_info: ReleaseInfo) -> str:
    clean_title = _clean_series_title(page_title)

    if release_info.season is not None:
        season_token = f"S{release_info.season:02d}"
    else:
        season_token = ""

    episode_token = ""
    if release_info.episode_min is not None:
        episode_min = release_info.episode_min
        episode_max = (
            release_info.episode_max
            if release_info.episode_max is not None
            else release_info.episode_min
        )
        episode_token = f"E{episode_min:02d}"
        if episode_max != episode_min:
            episode_token += f"-{episode_max:02d}"

    title_core = clean_title.strip().replace(" ", ".")
    if season_token:
        title_core += f".{season_token}{episode_token}"
    elif episode_token:
        title_core += f".{episode_token}"

    episode_title = _clean_episode_title(release_info.episode_title)
    if episode_title:
        title_core += f".{episode_title.replace(' ', '.')}"

    parts = [title_core]

    if release_info.season_part:
        part_string = f"Part.{release_info.season_part}"
        if part_string not in title_core:
            parts.append(part_string)

    audio_prefix = _derive_audio_prefix(release_info.audio_langs)
    if audio_prefix:
        parts.append(audio_prefix)

    resolved_subtitle_tokens = subtitle_tokens(release_info.subtitle_langs)
    if resolved_subtitle_tokens:
        parts.extend(resolved_subtitle_tokens)

    if release_info.audio:
        parts.append(release_info.audio)

    if release_info.resolution:
        parts.append(release_info.resolution)

    if release_info.source:
        parts.append(release_info.source)

    if release_info.video:
        parts.append(release_info.video)

    title = ".".join(parts)
    if release_info.release_group:
        title += f"-{release_info.release_group}"

    return inject_subtitle_tokens_in_title(title, release_info.subtitle_langs)


def _derive_audio_prefix(audio_langs: List[str]) -> str:
    if len(audio_langs) > 2 and "German" in audio_langs:
        return "German.ML"
    if len(audio_langs) == 2 and "German" in audio_langs:
        return "German.DL"
    if len(audio_langs) == 1 and "German" in audio_langs:
        return "German"
    if audio_langs:
        return audio_langs[0]
    return ""


def _clean_series_title(page_title: str) -> str:
    clean_title = str(page_title or "").strip()

    clean_title = re.sub(r"^\[[^\]]+\]\s*", "", clean_title)
    clean_title = re.sub(r"\s*\(\d{4}\)\s*$", "", clean_title)
    clean_title = re.sub(r"\s*\([^)]*\)\s*$", "", clean_title)
    clean_title = re.sub(
        r"(?i)\b(?:Season|Staffel)\s*\.?\s*\d+\b|\bR\d+\b", "", clean_title
    )
    clean_title = re.sub(r"\s*[-:|]\s*$", "", clean_title).strip()
    clean_title = re.sub(r"\s{2,}", " ", clean_title)

    return clean_title


def _clean_episode_title(episode_title: Optional[str]) -> str:
    cleaned = str(episode_title or "").strip()
    cleaned = re.sub(r"^[\s\-:|/._]+", "", cleaned)
    cleaned = re.sub(r"[\s\-:|/._]+$", "", cleaned)
    cleaned = re.sub(r"\s{2,}", " ", cleaned)
    return cleaned



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from abc import ABC, abstractmethod

from quasarr.providers.notifications.helpers.notification_message import (
    AbstractNotificationEntry,
    NotificationFactsEntry,
    NotificationLinkEntry,
    NotificationTextEntry,
    NotificationValueEntry,
)


class AbstractNotificationFormatter(ABC):
    def render_entry(self, entry: AbstractNotificationEntry):
        if isinstance(entry, NotificationTextEntry):
            return self.render_text_entry(entry)
        if isinstance(entry, NotificationLinkEntry):
            return self.render_link_entry(entry)
        if isinstance(entry, NotificationFactsEntry):
            return self.render_facts_entry(entry)
        if isinstance(entry, NotificationValueEntry):
            return self.render_value_entry(entry)
        raise TypeError(f"Unsupported notification entry type: {type(entry).__name__}")

    def render_entries(self, entries: tuple[AbstractNotificationEntry, ...]):
        return [self.render_entry(entry) for entry in entries]

    @abstractmethod
    def render_text_entry(self, entry: NotificationTextEntry):
        pass

    @abstractmethod
    def render_link_entry(self, entry: NotificationLinkEntry):
        pass

    @abstractmethod
    def render_facts_entry(self, entry: NotificationFactsEntry):
        pass

    @abstractmethod
    def render_value_entry(self, entry: NotificationValueEntry):
        pass



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re

from bottle import request, response

from quasarr.constants import NOTIFICATION_PROVIDERS, QUASARR_AVATAR
from quasarr.providers.log import info
from quasarr.providers.notifications.helpers.message_builder import (
    build_notification_message,
)
from quasarr.providers.notifications.helpers.notification_types import (
    NotificationType,
    get_user_configurable_notification_types,
)
from quasarr.storage.config import Config
from quasarr.storage.sqlite_database import DataBase

NOTIFICATION_TYPES = get_user_configurable_notification_types()
NOTIFICATION_SETTINGS_TABLE = "notification_settings"


def _notification_toggle_key(provider, notification_type):
    return f"{provider}_{notification_type.value}"


def _notification_silent_key(provider, notification_type):
    return f"{provider}_{notification_type.value}_silent"


def _coerce_bool(value, default=False):
    if isinstance(value, bool):
        return value
    if value is None:
        return default
    if isinstance(value, str):
        lowered = value.strip().lower()
        if lowered in {"1", "true", "yes", "on"}:
            return True
        if lowered in {"0", "false", "no", "off"}:
            return False
    return bool(value)


def _read_notification_setting(notification_settings_db, key, default=True):
    stored_value = notification_settings_db.retrieve(key)
    return _coerce_bool(stored_value, default)


def _read_notification_toggle(notification_settings_db, key):
    return _read_notification_setting(notification_settings_db, key, default=True)


def _read_notification_silent(notification_settings_db, key):
    return _read_notification_setting(notification_settings_db, key, default=False)


def _read_notification_settings():
    notification_config = Config("Notifications")
    notification_settings_db = DataBase(NOTIFICATION_SETTINGS_TABLE)

    settings = {
        "discord_webhook": notification_config.get("discord_webhook") or "",
        "telegram_bot_token": notification_config.get("telegram_bot_token") or "",
        "telegram_chat_id": notification_config.get("telegram_chat_id") or "",
        "toggles": {"discord": {}, "telegram": {}},
        "silent": {"discord": {}, "telegram": {}},
    }

    for provider in NOTIFICATION_PROVIDERS:
        for notification_type in NOTIFICATION_TYPES:
            case_key = notification_type.value
            settings["toggles"][provider][case_key] = _read_notification_toggle(
                notification_settings_db,
                _notification_toggle_key(provider, notification_type),
            )
            settings["silent"][provider][case_key] = _read_notification_silent(
                notification_settings_db,
                _notification_silent_key(provider, notification_type),
            )

    return settings


def _validate_notification_provider_credentials(
    discord_webhook,
    telegram_bot_token,
    telegram_chat_id,
):
    discord_webhook_pattern = r"^https://discord\.com/api/webhooks/\d+/[\w-]+$"
    telegram_token_pattern = r"^\d+:[A-Za-z0-9_-]{35,}$"

    if discord_webhook and not re.match(discord_webhook_pattern, discord_webhook):
        return "Invalid Discord Webhook URL"

    if telegram_bot_token or telegram_chat_id:
        if not telegram_bot_token or not telegram_chat_id:
            return "Telegram setup requires both bot token and chat ID"
        if not re.match(telegram_token_pattern, telegram_bot_token):
            return "Invalid Telegram bot token format"

    return None


def refresh_notification_settings(shared_state, backfill_defaults=False):
    del backfill_defaults

    settings = _read_notification_settings()

    shared_state.update("notification_settings", settings)

    return settings


def initialize_notification_settings(shared_state):
    return refresh_notification_settings(shared_state)


def get_notification_settings_data(shared_state):
    response.content_type = "application/json"
    settings = refresh_notification_settings(shared_state)
    return {"success": True, "settings": settings}


def save_notification_settings(shared_state):
    response.content_type = "application/json"

    data = request.json
    if not isinstance(data, dict):
        return {"success": False, "message": "Invalid JSON payload"}

    discord_webhook = str(data.get("discord_webhook", "")).strip()
    telegram_bot_token = str(data.get("telegram_bot_token", "")).strip()
    telegram_chat_id = str(data.get("telegram_chat_id", "")).strip()
    toggles = data.get("toggles") if isinstance(data.get("toggles"), dict) else {}
    silent = data.get("silent") if isinstance(data.get("silent"), dict) else {}

    validation_error = _validate_notification_provider_credentials(
        discord_webhook,
        telegram_bot_token,
        telegram_chat_id,
    )
    if validation_error:
        return {"success": False, "message": validation_error}

    notification_config = Config("Notifications")
    notification_config.save("discord_webhook", discord_webhook)
    notification_config.save("telegram_bot_token", telegram_bot_token)
    notification_config.save("telegram_chat_id", telegram_chat_id)
    notification_settings_db = DataBase(NOTIFICATION_SETTINGS_TABLE)

    for provider in NOTIFICATION_PROVIDERS:
        provider_toggles = toggles.get(provider)
        provider_silent = silent.get(provider)
        for notification_type in NOTIFICATION_TYPES:
            enabled_key = _notification_toggle_key(provider, notification_type)
            current_enabled = _read_notification_toggle(
                notification_settings_db,
                enabled_key,
            )
            if (
                isinstance(provider_toggles, dict)
                and notification_type.value in provider_toggles
            ):
                next_enabled = _coerce_bool(
                    provider_toggles.get(notification_type.value),
                    current_enabled,
                )
            else:
                next_enabled = current_enabled
            notification_settings_db.update_store(
                enabled_key, "true" if next_enabled else "false"
            )

            silent_key = _notification_silent_key(provider, notification_type)
            current_silent = _read_notification_silent(
                notification_settings_db,
                silent_key,
            )
            if (
                isinstance(provider_silent, dict)
                and notification_type.value in provider_silent
            ):
                next_silent = _coerce_bool(
                    provider_silent.get(notification_type.value),
                    current_silent,
                )
            else:
                next_silent = current_silent
            notification_settings_db.update_store(
                silent_key, "true" if next_silent else "false"
            )

    settings = refresh_notification_settings(shared_state)

    return {
        "success": True,
        "message": "Notification settings saved successfully",
        "settings": settings,
    }


def send_notification_test(shared_state):
    response.content_type = "application/json"

    data = request.json
    if not isinstance(data, dict):
        return {"success": False, "message": "Invalid JSON payload"}

    provider = str(data.get("provider", "")).strip().lower()
    if provider not in NOTIFICATION_PROVIDERS:
        return {"success": False, "message": "Unknown provider"}

    settings = refresh_notification_settings(shared_state)
    title = "Quasarr Notification Test"
    test_image_url = QUASARR_AVATAR

    if provider == "discord":
        if not settings["discord_webhook"]:
            return {
                "success": False,
                "message": "Discord webhook is not configured",
            }
        from quasarr.providers.notifications import discord

        message = build_notification_message(
            shared_state,
            title=title,
            case=NotificationType.TEST,
            details={"provider": "Discord"},
            image_url=test_image_url,
        )
        if message is None:
            return {"success": False, "message": "Failed to build Discord test message"}

        try:
            sent = discord.send(shared_state, message, silent=False)
        except Exception as e:
            info(f"Discord test notification error: {e}")
            return {
                "success": False,
                "message": f"Failed to send Discord test message: {e}",
            }
        if sent:
            return {"success": True, "message": "Discord test message sent"}
        return {
            "success": False,
            "message": "Failed to send Discord test message",
        }

    if not settings["telegram_bot_token"] or not settings["telegram_chat_id"]:
        return {
            "success": False,
            "message": "Telegram bot token and chat ID are required",
        }

    from quasarr.providers.notifications import telegram

    message = build_notification_message(
        shared_state,
        title=title,
        case=NotificationType.TEST,
        details={"provider": "Telegram"},
        image_url=test_image_url,
    )
    if message is None:
        return {"success": False, "message": "Failed to build Telegram test message"}

    try:
        sent = telegram.send(shared_state, message, silent=False)
        if not sent:
            telegram_destination = telegram.inspect_destination(shared_state)
    except Exception as e:
        info(f"Telegram test notification error: {e}")
        return {
            "success": False,
            "message": f"Failed to send Telegram test message: {e}",
        }
    if sent:
        return {"success": True, "message": "Telegram test message sent"}

    destination_message = telegram_destination.get("message")
    if destination_message:
        return {
            "success": False,
            "message": f"Failed to send Telegram test message: {destination_message}",
        }
    return {"success": False, "message": "Failed to send Telegram test message"}



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re

from bottle import Bottle, request, response

import quasarr.providers.web_server
from quasarr.constants import FALLBACK_USER_AGENT
from quasarr.providers.html_templates import (
    render_button,
    render_fail,
    render_form,
)
from quasarr.providers.log import info
from quasarr.providers.utils import check_flaresolverr
from quasarr.providers.web_server import Server
from quasarr.storage.config import Config
from quasarr.storage.setup.common import (
    add_no_cache_headers,
    render_reconnect_success,
    setup_auth,
)
from quasarr.storage.sqlite_database import DataBase


def save_flaresolverr_url(shared_state, is_setup=False):
    """Save FlareSolverr URL from web UI."""
    url = request.forms.get("url", "").strip()
    config = Config("FlareSolverr")

    if not url:
        config.save("url", "")
        DataBase("skip_flaresolverr").update_store("skipped", "true")
        shared_state.update("user_agent", FALLBACK_USER_AGENT)
        info("FlareSolverr URL cleared and setup skipped")

        if is_setup:
            quasarr.providers.web_server.temp_server_success = True

        return render_reconnect_success("FlareSolverr URL cleared.")

    if not url.startswith("http://") and not url.startswith("https://"):
        url = "http://" + url

    if not re.search(r"/v\d+$", url):
        return render_fail(
            "FlareSolverr URL must end with /v1 (or similar version path)."
        )

    if check_flaresolverr(shared_state, url):
        config.save("url", url)
        DataBase("skip_flaresolverr").delete("skipped")

        info(
            f'FlareSolverr connection successful. Using User-Agent: "{shared_state.values["user_agent"]}"'
        )
        info(f'FlareSolverr URL configured: "{url}"')

        if is_setup:
            quasarr.providers.web_server.temp_server_success = True

        return render_reconnect_success("FlareSolverr URL saved successfully!")
    return render_fail("Could not reach FlareSolverr!")


def get_flaresolverr_status_data(shared_state):
    """Return FlareSolverr configuration status."""
    response.content_type = "application/json"
    skip_db = DataBase("skip_flaresolverr")
    is_skipped = bool(skip_db.retrieve("skipped"))
    current_url = Config("FlareSolverr").get("url") or ""

    is_working = False
    if current_url and not is_skipped:
        is_working = check_flaresolverr(shared_state, current_url)

    return {"skipped": is_skipped, "url": current_url, "working": is_working}


def delete_skip_flaresolverr_preference():
    """Clear skip FlareSolverr preference."""
    response.content_type = "application/json"
    DataBase("skip_flaresolverr").delete("skipped")
    info("Skip FlareSolverr preference cleared")
    return {"success": True}


def flaresolverr_form_html(shared_state, is_setup=False):
    skip_db = DataBase("skip_flaresolverr")
    is_skipped = skip_db.retrieve("skipped")
    current_url = Config("FlareSolverr").get("url") or ""

    skip_indicator = ""
    if is_skipped and not is_setup:
        skip_indicator = """
        <div class="skip-indicator" style="margin-bottom:1rem; padding:0.75rem; background:var(--code-bg, #f8f9fa); border-radius:0.25rem; font-size:0.875rem;">
            <span style="color:#dc3545;">⚠️ FlareSolverr setup was skipped</span>
            <p style="margin:0.5rem 0 0 0; font-size:0.75rem; color:var(--secondary, #6c757d);">
                Some sites (like AL) won't work until FlareSolverr is configured.
            </p>
        </div>
        """

    form_content = f'''
    {skip_indicator}
    <span><a href="https://github.com/FlareSolverr/FlareSolverr?tab=readme-ov-file#installation" target="_blank">FlareSolverr</a>
    must be running and reachable to Quasarr for some sites to work.</span><br><br>
    <label for="url">FlareSolverr URL</label>
    <input type="text" id="url" name="url" placeholder="http://192.168.0.1:8191/v1" value="{current_url}"><br>
    '''

    buttons = render_button("Save", "primary", {"type": "submit", "id": "submitBtn"})
    extra_js = ""

    if is_setup:
        buttons += ' <button type="button" class="btn-warning" id="skipBtn" onclick="skipFlaresolverr()">Skip for now</button>'
        extra_js = """
        function skipFlaresolverr() {
            if (formSubmitted) return;
            formSubmitted = true;
            var skipBtn = document.getElementById('skipBtn');
            var submitBtn = document.getElementById('submitBtn');
            if (skipBtn) { skipBtn.disabled = true; skipBtn.textContent = 'Skipping...'; }
            if (submitBtn) { submitBtn.disabled = true; }

            quasarrApiFetch('/api/flaresolverr/skip', { method: 'POST' })
            .then(response => {
                if (response.ok) {
                    window.location.href = '/skip-success';
                } else {
                    showModal('Error', 'Failed to skip FlareSolverr setup');
                    formSubmitted = false;
                    if (skipBtn) { skipBtn.disabled = false; skipBtn.textContent = 'Skip for now'; }
                    if (submitBtn) { submitBtn.disabled = false; }
                }
            })
            .catch(error => {
                showModal('Error', 'Error: ' + error.message);
                formSubmitted = false;
                if (skipBtn) { skipBtn.disabled = false; skipBtn.textContent = 'Skip for now'; }
                if (submitBtn) { submitBtn.disabled = false; }
            });
        }
        """

    form_html = f"""
    <style>
        .button-row {{
            display: flex;
            gap: 0.75rem;
            justify-content: center;
            flex-wrap: wrap;
            margin-top: 1rem;
        }}
        .btn-warning {{
            background-color: #ffc107;
            color: #212529;
            border: 1.5px solid #d39e00;
            padding: 0.5rem 1rem;
            font-size: 1rem;
            border-radius: 0.5rem;
            font-weight: 500;
            cursor: pointer;
        }}
        .btn-warning:hover {{
            background-color: #e0a800;
            border-color: #c69500;
        }}
    </style>
    <form action="/api/flaresolverr" method="post" onsubmit="return handleSubmit(this)">
        {form_content}
        <div class="button-row">
            {buttons}
        </div>
    </form>
    <script>
    var formSubmitted = false;
    function handleSubmit(form) {{
        if (formSubmitted) return false;
        formSubmitted = true;
        var btn = document.getElementById('submitBtn');
        if (btn) {{ btn.disabled = true; btn.textContent = 'Saving...'; }}
        var skipBtn = document.getElementById('skipBtn');
        if (skipBtn) {{ skipBtn.disabled = true; }}
        return true;
    }}
    {extra_js}
    </script>
    """

    if not is_setup:
        form_html += f"""<p>{render_button("Back", "secondary", {"onclick": "location.href='/'"})}</p>"""

    return form_html


def flaresolverr_config(shared_state):
    app = Bottle()
    add_no_cache_headers(app)
    setup_auth(app)

    @app.get("/")
    def url_form():
        return render_form(
            "Set FlareSolverr URL", flaresolverr_form_html(shared_state, is_setup=True)
        )

    @app.get("/skip-success")
    def skip_success():
        return render_reconnect_success(
            "FlareSolverr setup skipped. Some sites (like AL) won't work. You can configure it later in the web UI."
        )

    @app.post("/api/flaresolverr/skip")
    def skip_flaresolverr():
        DataBase("skip_flaresolverr").update_store("skipped", "true")
        shared_state.update("user_agent", FALLBACK_USER_AGENT)
        info("FlareSolverr setup skipped by user choice")
        quasarr.providers.web_server.temp_server_success = True
        return {"success": True}

    @app.post("/api/flaresolverr")
    def set_flaresolverr_url_route():
        return save_flaresolverr_url(shared_state, is_setup=True)

    info(
        '"flaresolverr" URL is required for some sites (like AL). '
        f'Starting web server for config at: "{shared_state.values["external_address"]}".'
    )
    info("Please enter your FlareSolverr URL now, or skip to allow Quasarr to launch!")
    quasarr.providers.web_server.temp_server_success = False
    return Server(
        app, listen="0.0.0.0", port=shared_state.values["port"]
    ).serve_temporarily()



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from bottle import response

import quasarr.providers.html_images as images
from quasarr.providers.auth import add_auth_hook, add_auth_routes
from quasarr.providers.html_templates import (
    render_button,
    render_centered_html,
)


def render_reconnect_success(message, countdown_seconds=3):
    """Render a success page that waits, then polls until the server is back online."""
    button_html = render_button(
        f"Continuing in {countdown_seconds}...",
        "secondary",
        {"id": "reconnectBtn", "disabled": "true"},
    )

    script = f"""
        <script>
            var remaining = {countdown_seconds};
            var btn = document.getElementById('reconnectBtn');

            var interval = setInterval(function() {{
                remaining--;
                btn.innerText = 'Continuing in ' + remaining + '...';
                if (remaining <= 0) {{
                    clearInterval(interval);
                    btn.innerText = 'Reconnecting...';
                    tryReconnect();
                }}
            }}, 1000);

            function tryReconnect() {{
                var attempts = 0;
                function attempt() {{
                    attempts++;
                    fetch('/', {{ method: 'HEAD', cache: 'no-store' }})
                    .then(function(response) {{
                        if (response.ok) {{
                            btn.innerText = 'Connected! Reloading...';
                            btn.className = 'btn-primary';
                            setTimeout(function() {{ window.location.href = '/'; }}, 500);
                        }} else {{
                            scheduleRetry();
                        }}
                    }})
                    .catch(function() {{
                        scheduleRetry();
                    }});
                }}
                function scheduleRetry() {{
                    btn.innerText = 'Reconnecting... (attempt ' + attempts + ')';
                    setTimeout(attempt, 1000);
                }}
                attempt();
            }}
        </script>
    """

    content = f'''<h1><img src="{images.logo}" type="image/webp" alt="Quasarr logo" class="logo"/>Quasarr</h1>
    <h2>✅ Success</h2>
    <p>{message}</p>
    {button_html}
    {script}
    '''
    return render_centered_html(content)


def add_no_cache_headers(app):
    """Add hooks to prevent browser caching of setup pages."""

    @app.hook("after_request")
    def set_no_cache():
        response.set_header("Cache-Control", "no-cache, no-store, must-revalidate")
        response.set_header("Pragma", "no-cache")
        response.set_header("Expires", "0")


def setup_auth(app):
    """Add authentication to setup app if enabled."""
    add_auth_routes(app)
    add_auth_hook(app)


__all__ = [
    "add_no_cache_headers",
    "render_reconnect_success",
    "setup_auth",
]



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from bottle import Bottle, request, response

import quasarr.providers.web_server
from quasarr.providers.html_templates import (
    render_button,
    render_fail,
    render_form,
)
from quasarr.providers.log import info
from quasarr.providers.web_server import Server
from quasarr.storage.config import Config
from quasarr.storage.setup.common import (
    add_no_cache_headers,
    render_reconnect_success,
    setup_auth,
)


def verify_jdownloader_credentials(shared_state):
    """Verify JDownloader credentials and return devices."""
    response.content_type = "application/json"
    try:
        data = request.json
        username = data.get("user")
        password = data.get("pass")

        devices = shared_state.get_devices(username, password)
        device_names = []
        if devices:
            for device in devices:
                device_names.append(device["name"])

        if device_names:
            return {"success": True, "devices": device_names}
        return {
            "success": False,
            "message": "No devices found or invalid credentials",
        }
    except Exception as e:
        return {"success": False, "message": str(e)}


def save_jdownloader_settings(shared_state, is_setup=False):
    """Save JDownloader settings."""
    if request.json:
        username = request.json.get("user")
        password = request.json.get("pass")
        device = request.json.get("device")
    else:
        username = request.forms.get("user")
        password = request.forms.get("pass")
        device = request.forms.get("device")

    if username and password and device:
        if shared_state.set_device(username, password, device):
            config = Config("JDownloader")
            config.save("user", username)
            config.save("password", password)
            config.save("device", device)

            if is_setup:
                quasarr.providers.web_server.temp_server_success = True
                return render_reconnect_success("Credentials set")

            response.content_type = "application/json"
            return {
                "success": True,
                "message": "JDownloader configured successfully",
            }

        if is_setup:
            return render_fail("Could not connect to selected device!")

        response.content_type = "application/json"
        return {
            "success": False,
            "message": "Could not connect to selected device",
        }

    if is_setup:
        return render_fail("Could not set credentials!")

    response.content_type = "application/json"
    return {"success": False, "message": "Missing required fields"}


def jdownloader_config(shared_state):
    app = Bottle()
    add_no_cache_headers(app)
    setup_auth(app)

    @app.get("/")
    def jd_form():
        verify_form_html = f"""
        <span>If required register account at: <a href="https://my.jdownloader.org/login.html#register" target="_blank">
        my.jdownloader.org</a>!</span><br>

        <p><strong>JDownloader must be running and connected to My JDownloader!</strong></p><br>

        <form id="verifyForm" action="/api/verify_jdownloader" method="post">
            <label for="user">E-Mail</label>
            <input type="text" id="user" name="user" placeholder="user@example.org" autocorrect="off"><br>
            <label for="pass">Password</label>
            <input type="password" id="pass" name="pass" placeholder="Password"><br>
            {
            render_button(
                "Verify Credentials",
                "secondary",
                {
                    "id": "verifyButton",
                    "type": "button",
                    "onclick": "verifyCredentials()",
                },
            )
        }
        </form>

        <p>Some JDownloader settings will be enforced by Quasarr on startup.</p>

        <form action="/api/store_jdownloader" method="post" id="deviceForm" style="display: none;" onsubmit="return handleStoreSubmit(this)">
            <input type="hidden" id="hiddenUser" name="user">
            <input type="hidden" id="hiddenPass" name="pass">
            <label for="device">JDownloader</label>
            <select id="device" name="device"></select><br>
            {render_button("Save", "primary", {"type": "submit", "id": "storeBtn"})}
        </form>
        <p><strong>Saving may take a while!</strong></p><br>
        """

        verify_script = """
        <script>
        var verifyInProgress = false;
        var storeSubmitted = false;
        function verifyCredentials() {
            if (verifyInProgress) return;
            verifyInProgress = true;
            var btn = document.getElementById('verifyButton');
            if (btn) { btn.disabled = true; btn.textContent = 'Verifying...'; }

            var user = document.getElementById('user').value;
            var pass = document.getElementById('pass').value;
            quasarrApiFetch('/api/verify_jdownloader', {
                method: 'POST',
                headers: {
                    'Content-Type': 'application/json',
                },
                body: JSON.stringify({user: user, pass: pass}),
            })
            .then(response => response.json())
            .then(data => {
                if (data.success) {
                    var select = document.getElementById('device');
                    data.devices.forEach(device => {
                        var opt = document.createElement('option');
                        opt.value = device;
                        opt.innerHTML = device;
                        select.appendChild(opt);
                    });
                    document.getElementById('hiddenUser').value = document.getElementById('user').value;
                    document.getElementById('hiddenPass').value = document.getElementById('pass').value;
                    document.getElementById("verifyButton").style.display = "none";
                    document.getElementById('deviceForm').style.display = 'block';
                } else {
                    showModal('Error', 'Error! Please check your Credentials.');
                    verifyInProgress = false;
                    if (btn) { btn.disabled = false; btn.textContent = 'Verify Credentials'; }
                }
            })
            .catch((error) => {
                console.error('Error:', error);
                verifyInProgress = false;
                if (btn) { btn.disabled = false; btn.textContent = 'Verify Credentials'; }
            });
        }
        function handleStoreSubmit(form) {
            if (storeSubmitted) return false;
            storeSubmitted = true;
            var btn = document.getElementById('storeBtn');
            if (btn) { btn.disabled = true; btn.textContent = 'Saving...'; }
            return true;
        }
        </script>
        """
        return render_form(
            "Set your credentials for My JDownloader", verify_form_html, verify_script
        )

    @app.post("/api/verify_jdownloader")
    def verify_jdownloader():
        return verify_jdownloader_credentials(shared_state)

    @app.post("/api/store_jdownloader")
    def store_jdownloader():
        return save_jdownloader_settings(shared_state, is_setup=True)

    info(
        f"My-JDownloader-Credentials not set. "
        f'Starting web server for config at: "{shared_state.values["external_address"]}".'
    )
    info("If needed register here: 'https://my.jdownloader.org/login.html#register'")
    info("Please set your credentials now, to allow Quasarr to launch!")
    quasarr.providers.web_server.temp_server_success = False
    return Server(
        app, listen="0.0.0.0", port=shared_state.values["port"]
    ).serve_temporarily()



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re
from datetime import datetime, timedelta
from functools import lru_cache
from urllib.parse import urlparse

import requests
from bs4 import BeautifulSoup

from quasarr.constants import (
    DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
    SHARE_HOSTERS_LOWERCASE,
)
from quasarr.downloads.sources.helpers.abstract_source import AbstractDownloadSource
from quasarr.downloads.sources.helpers.anime_title import (
    ReleaseInfo,
    guess_release_title,
)
from quasarr.providers.hostname_issues import clear_hostname_issue, mark_hostname_issue
from quasarr.providers.log import info
from quasarr.providers.utils import sanitize_string
from quasarr.providers.xem_metadata import get_all_season_names

_SEASON_EP_REGEX = re.compile(
    r"\bS(?:eason)?\s*0*(\d{1,3})\s*[.\-_ ]?(?:E|EP|Episode)\s*0*(\d{1,4})(?:\s*[-+]\s*0*(\d{1,4}))?\b",
    re.I,
)
_SEASON_DASH_EP_REGEX = re.compile(
    r"\bS(?:eason)?\s*0*(\d{1,3})\s*-\s*0*(\d{1,4})(?:\s*[-+]\s*0*(\d{1,4}))?\b",
    re.I,
)
_SEASON_ONLY_REGEX = re.compile(
    r"\b(?:Season|Staffel)\s*0*(\d{1,3})\b|\b0*(\d{1,3})(?:st|nd|rd|th)\s+Season\b|\bS0*(\d{1,3})\b",
    re.I,
)
_ABSOLUTE_EP_REGEX = re.compile(
    r"\b(?:EP?|Episode)\s*0*(\d{1,4})(?:\s*[-+]\s*0*(\d{1,4}))?\b", re.I
)
_ABSOLUTE_RANGE_REGEX = re.compile(r"[\[(]0*(\d{1,4})\s*[-+]\s*0*(\d{1,4})[\])]", re.I)
_TITLE_DASH_EP_REGEX = re.compile(
    r"\s-\s0*(\d{1,4})(?:\s*[-+]\s*0*(\d{1,4}))?(?=\s[\[(]|\s(?:WEB|BD|BluRay|CR|DVD|HDTV|NF|AMZN|BILI|IQIYI)|$)",
    re.I,
)
_STANDALONE_EP_REGEX = re.compile(
    r"(?:\s-\s|\s)0*(\d{1,4})(?:v\d+)?(?:\s*[-+]\s*0*(\d{1,4}))?(?=\s*(?:END\b|FINAL\b|V\d+\b|[\[(]|\||$))",
    re.I,
)
_PART_REGEX = re.compile(r"\b(?:Part|Teil)\s+(\d+)\b", re.I)
_EPISODE_TITLE_STOP_REGEX = re.compile(
    r"""
    (?=
        \s*(?:\(|\[)?
        (?:
            \d{3,4}p|4k|
            cr|nf|amzn|bili|iqiyi|
            web(?:-dl|dl|rip)?|blu-?ray|bluray|bd|hdtv|tvrip|
            aac(?:\d(?:\.\d)?)?|ddp\d(?:\.\d)?|dd\d(?:\.\d)?|
            eac3|ac3|flac|opus|mp3|pcm|dts|
            hevc|avc|av1|x26[45]|h[\s.]?26[45]|
            10bit|multi(?:[- ]?sub(?:s)?)|dual(?:[- ]audio)?|
            english\s+dub|german\s+dub|weekly|batch
        )\b
    )
    """,
    re.I | re.X,
)
_SOURCE_TAG_PATTERNS = (
    (re.compile(r"\bcr\b", re.I), "CR"),
    (re.compile(r"\bnf\b", re.I), "NF"),
    (re.compile(r"\bamzn\b", re.I), "AMZN"),
    (re.compile(r"\bbili\b", re.I), "BILI"),
    (re.compile(r"\biqiyi\b", re.I), "IQIYI"),
)
_AUDIO_CODEC_PATTERNS = (
    (re.compile(r"\bDDP\d(?:\.\d)?\b", re.I), lambda m: m.group(0).upper()),
    (re.compile(r"\bDD\d(?:\.\d)?\b", re.I), lambda m: m.group(0).upper()),
    (re.compile(r"\bAAC\d(?:\.\d)?\b", re.I), lambda m: m.group(0).upper()),
    (re.compile(r"\bEAC3\b", re.I), lambda m: "EAC3"),
    (re.compile(r"\bAC3\b", re.I), lambda m: "AC3"),
    (re.compile(r"\bFLAC\b", re.I), lambda m: "FLAC"),
    (re.compile(r"\bOPUS\b", re.I), lambda m: "Opus"),
    (re.compile(r"\bMP3\b", re.I), lambda m: "MP3"),
    (re.compile(r"\bPCM\b", re.I), lambda m: "PCM"),
    (re.compile(r"\bDTS\b", re.I), lambda m: "DTS"),
    (re.compile(r"\bAAC\b", re.I), lambda m: "AAC"),
)
_DIRECT_DOWNLOAD_MIRROR_ALIASES = {
    "down": "mdiaload",
}
_JAPANESE_SEASON_NAMES = {
    "ni": 2,
    "san": 3,
    "yon": 4,
    "go": 5,
    "roku": 6,
    "nana": 7,
    "shichi": 7,
    "hachi": 8,
    "kyuu": 9,
    "ku": 9,
    "juu": 10,
}


class Source(AbstractDownloadSource):
    initials = "at"

    def get_download_links(self, shared_state, url, mirrors, title, password):
        headers = {"User-Agent": shared_state.values["user_agent"]}

        try:
            response = requests.get(
                url,
                headers=headers,
                timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
            )
            response.raise_for_status()
        except Exception as e:
            info(f"Could not load release page for {title}: {e}")
            mark_hostname_issue(Source.initials, "download", str(e) or "Download error")
            return {"links": [], "imdb_id": None}

        try:
            soup = BeautifulSoup(response.text, "html.parser")
            raw_title = _extract_release_title(soup)
            series_title = _extract_series_title_from_release_page(soup)
            subtitle_langs = _extract_subtitle_langs_from_release_page(soup)
            release_info = _build_release_info_from_title(raw_title, subtitle_langs)
            page_title, release_info = _resolve_title_context(
                series_title
                or _extract_series_title_from_raw_title(raw_title)
                or raw_title,
                raw_title,
                release_info,
            )
            resolved_title = guess_release_title(page_title, release_info)

            links = []
            for cell in _iter_link_cells(soup):
                links.extend(
                    _extract_direct_links_from_anchors(cell.find_all("a"), mirrors)
                )

            links = _dedupe_links(links)
            if links:
                clear_hostname_issue(Source.initials)

            return {"links": links, "imdb_id": None, "title": resolved_title}
        except Exception as e:
            info(f"Could not parse release page for {title}: {e}")
            mark_hostname_issue(Source.initials, "download", str(e) or "Download error")
            return {"links": [], "imdb_id": None}


def _iter_link_cells(soup):
    for row in soup.find_all("tr"):
        th = row.find("th")
        td = row.find("td")
        if not th or not td:
            continue

        label = th.get_text(" ", strip=True).lower()
        if label in {"source links", "extractions"}:
            yield td


def _extract_direct_links_from_anchors(anchors, mirrors=None):
    requested_mirrors = {
        _normalize_requested_mirror(mirror) for mirror in (mirrors or []) if mirror
    }
    links = []

    for anchor in anchors:
        href = (anchor.get("href") or "").strip()
        mirror = _derive_supported_mirror_from_url(href)
        if not mirror:
            continue

        if requested_mirrors and mirror not in requested_mirrors:
            continue

        links.append([href, mirror])

    return _dedupe_links(links)


def _dedupe_links(links):
    deduped = []
    seen = set()

    for href, mirror in links:
        if href in seen:
            continue
        seen.add(href)
        deduped.append([href, mirror])

    return deduped


def _normalize_requested_mirror(mirror_name):
    normalized = str(mirror_name or "").lower().strip()

    if "://" in normalized:
        parsed = urlparse(normalized)
        normalized = parsed.netloc or parsed.path

    if normalized.startswith("www."):
        normalized = normalized[4:]

    normalized = normalized.split("/", 1)[0]
    normalized = normalized.split(":", 1)[0]
    if "." in normalized:
        normalized = normalized.split(".", 1)[0]

    return _DIRECT_DOWNLOAD_MIRROR_ALIASES.get(normalized, normalized)


def _derive_supported_mirror_from_url(url):
    try:
        hostname = (urlparse(url).hostname or "").lower()
    except Exception:
        return None

    if not hostname:
        return None

    if hostname.startswith("www."):
        hostname = hostname[4:]

    provider = hostname.split(".", 1)[0]
    provider = _DIRECT_DOWNLOAD_MIRROR_ALIASES.get(provider, provider)
    if provider in SHARE_HOSTERS_LOWERCASE:
        return provider
    return None


def _extract_release_title(soup):
    title_node = soup.select_one("#title")
    return title_node.get_text(" ", strip=True) if title_node else ""


def _extract_series_title_from_release_page(soup):
    for row in soup.find_all("tr"):
        th = row.find("th")
        td = row.find("td")
        if not th or not td:
            continue

        if not th.get_text(" ", strip=True).lower().startswith("series"):
            continue

        anchor = td.find("a")
        if anchor:
            return _clean_series_title(
                anchor.get("title") or anchor.get_text(" ", strip=True)
            )

    breadcrumb_links = soup.select("#nav_bc a")
    if len(breadcrumb_links) >= 2:
        return _clean_series_title(
            breadcrumb_links[-2].get("title")
            or breadcrumb_links[-2].get_text(" ", strip=True)
        )

    return ""


def _extract_series_title_from_listing_entry(entry):
    anchor = entry.select_one("span.serieslink a")
    if not anchor:
        return ""

    return _clean_series_title(anchor.get("title") or anchor.get_text(" ", strip=True))


def _extract_subtitle_langs_from_release_page(soup):
    for row in soup.find_all("tr"):
        th = row.find("th")
        td = row.find("td")
        if not th or not td:
            continue

        if th.get_text(" ", strip=True).lower() != "extractions":
            continue

        return _extract_subtitle_langs_from_links(td)

    return []


def _extract_subtitle_langs_from_links(container):
    subtitle_langs = []

    for anchor in container.find_all("a", href=True):
        href = (anchor.get("href") or "").strip().lower()
        if "_track" not in href:
            continue

        text = anchor.get_text(" ", strip=True)
        for match in re.findall(r"\[([^\]]+)\]", text):
            for token in re.split(r"[\s,/_-]+", match):
                _add_subtitle_lang(token, subtitle_langs)

        for token in re.split(r"[\s,/_-]+", re.sub(r"\[[^\]]+\]", " ", text)):
            _add_subtitle_lang(token, subtitle_langs)

    return subtitle_langs


def _add_subtitle_lang(token, subtitle_langs):
    token = token.strip().lower()
    if token in {"", "ass", "ssa", "srt", "sub", "subs", "forced", "signs", "cr"}:
        return

    if token not in subtitle_langs:
        subtitle_langs.append(token)


def _clean_series_title(series_title):
    cleaned = str(series_title or "").replace("...", "").strip()
    cleaned = re.sub(r"\s*\(\d{4}\)\s*$", "", cleaned)
    cleaned = re.sub(r"\s{2,}", " ", cleaned)
    return cleaned.strip()


def _strip_file_extension(title):
    return re.sub(
        r"\.(?:mkv|mp4|avi|m2ts|mov|zip|7z)$", "", str(title or ""), flags=re.I
    )


def _build_release_info_from_title(raw_title, subtitle_langs=None, forced_season=None):
    raw_title = _strip_file_extension(raw_title)
    subtitle_langs = list(subtitle_langs or [])
    title_lower = raw_title.lower()

    season = _extract_season_number(raw_title)
    if season is None and forced_season is not None:
        season = int(forced_season)

    episode_min, episode_max = _extract_episode_range(raw_title, season)

    return ReleaseInfo(
        release_title=None,
        audio_langs=_extract_audio_langs(title_lower),
        subtitle_langs=subtitle_langs,
        episode_title=_extract_episode_title(raw_title),
        resolution=_extract_resolution(title_lower),
        audio=_extract_audio_codec(raw_title),
        video=_extract_video_codec(raw_title),
        source=_extract_source(title_lower),
        release_group=_extract_release_group(raw_title),
        season_part=_extract_season_part(raw_title),
        season=season,
        episode_min=episode_min,
        episode_max=episode_max,
    )


def _extract_season_number(raw_title):
    season_ep_match = _SEASON_EP_REGEX.search(raw_title)
    if season_ep_match:
        return int(season_ep_match.group(1))

    season_dash_match = _SEASON_DASH_EP_REGEX.search(raw_title)
    if season_dash_match:
        return int(season_dash_match.group(1))

    season_match = _SEASON_ONLY_REGEX.search(raw_title)
    if season_match:
        for value in season_match.groups():
            if value:
                return int(value)

    return None


def _extract_season_hint_from_title(title):
    normalized_title = str(title or "")
    if not normalized_title:
        return None

    explicit_match = _SEASON_ONLY_REGEX.search(normalized_title)
    if explicit_match:
        for value in explicit_match.groups():
            if value:
                return int(value)

    ordinal_match = re.search(
        r"\b0*(\d{1,2})(?:st|nd|rd|th)\s+Season\b", normalized_title, re.I
    )
    if ordinal_match:
        return int(ordinal_match.group(1))

    japanese_match = re.search(r"\b([A-Za-z]+)\s+no\s+Shou\b", normalized_title, re.I)
    if japanese_match:
        mapped = _JAPANESE_SEASON_NAMES.get(japanese_match.group(1).lower())
        if mapped:
            return mapped

    return None


def _resolve_trailing_season_context(page_title):
    trailing_match = re.search(
        r"^(?P<base>.*\S)\s+(?P<season>[2-9]|1\d)\s*$",
        str(page_title or "").strip(),
    )
    if not trailing_match:
        return None, None

    base_title = _clean_series_title(trailing_match.group("base"))
    if not base_title:
        return None, None

    season = int(trailing_match.group("season"))
    season_names = _get_cached_season_names(base_title)
    if not season_names:
        return None, None

    known_seasons = {
        int(season_key) for season_key in season_names if str(season_key).isdigit()
    }
    if season not in known_seasons:
        return None, None

    canonical_title = season_names.get("all") or base_title
    if isinstance(canonical_title, list):
        canonical_title = canonical_title[0] if canonical_title else base_title

    return str(canonical_title or base_title), season


def _extract_episode_range(raw_title, season):
    season_ep_match = _SEASON_EP_REGEX.search(raw_title)
    if season_ep_match:
        episode_min = int(season_ep_match.group(2))
        episode_max = int(season_ep_match.group(3) or episode_min)
        return episode_min, episode_max

    season_dash_match = _SEASON_DASH_EP_REGEX.search(raw_title)
    if season_dash_match:
        episode_min = int(season_dash_match.group(2))
        episode_max = int(season_dash_match.group(3) or episode_min)
        return episode_min, episode_max

    absolute_match = _ABSOLUTE_EP_REGEX.search(raw_title)
    if absolute_match:
        episode_min = int(absolute_match.group(1))
        episode_max = int(absolute_match.group(2) or episode_min)
        return episode_min, episode_max

    absolute_range_match = _ABSOLUTE_RANGE_REGEX.search(raw_title)
    if absolute_range_match:
        return int(absolute_range_match.group(1)), int(absolute_range_match.group(2))

    dash_match = _TITLE_DASH_EP_REGEX.search(raw_title)
    if dash_match:
        episode_min = int(dash_match.group(1))
        episode_max = int(dash_match.group(2) or episode_min)
        return episode_min, episode_max

    standalone_match = _STANDALONE_EP_REGEX.search(raw_title)
    if standalone_match:
        episode_min = int(standalone_match.group(1))
        episode_max = int(standalone_match.group(2) or episode_min)
        return episode_min, episode_max

    if season is None:
        plain_range_match = re.search(r"\b(\d{2,4})\s*[-+]\s*(\d{2,4})\b", raw_title)
        if plain_range_match:
            return int(plain_range_match.group(1)), int(plain_range_match.group(2))

    return None, None


def _extract_series_title_from_raw_title(raw_title):
    title = _strip_file_extension(raw_title)
    title = re.sub(r"^\[[^\]]+\]\s*", "", title).strip()
    title = title.split("|", 1)[0].strip()
    title = re.sub(r"\[[^\]]+\]", " ", title)
    title = re.sub(
        r"\((?:[^)]*(?:multi|subs|audio|weekly|batch|eng dub|english dub|ger dub|german dub)[^)]*)\)",
        " ",
        title,
        flags=re.I,
    )

    title = re.sub(r"(?i)\b(?:Part|Teil)\s+\d+\b", "", title)

    match = (
        _SEASON_EP_REGEX.search(title)
        or _SEASON_DASH_EP_REGEX.search(title)
        or _ABSOLUTE_EP_REGEX.search(title)
        or _ABSOLUTE_RANGE_REGEX.search(title)
        or _TITLE_DASH_EP_REGEX.search(title)
        or _STANDALONE_EP_REGEX.search(title)
    )
    if match:
        title = title[: match.start()]

    title = re.sub(r"\s*\([^)]*\)\s*$", "", title)
    title = re.sub(r"\s*[-:|/._]+\s*$", "", title)
    title = re.sub(r"\s{2,}", " ", title)
    return title.strip()


def _extract_episode_title(raw_title):
    title = _strip_file_extension(raw_title)
    match = (
        _SEASON_EP_REGEX.search(title)
        or _SEASON_DASH_EP_REGEX.search(title)
        or _ABSOLUTE_EP_REGEX.search(title)
        or _ABSOLUTE_RANGE_REGEX.search(title)
        or _TITLE_DASH_EP_REGEX.search(title)
    )
    if not match:
        return None

    tail = title[match.end() :]
    if tail.lstrip().startswith(("[", "(")):
        return None
    tail = tail.split("|", 1)[0]
    tail = tail.split(" / ", 1)[0]
    tail = re.sub(r"^\s*[-:|/._]+\s*", "", tail)
    tail = re.sub(r"\[[^\]]+\]", " ", tail)
    tail = re.sub(r"^\s*\([^)]*\)\s*", "", tail)
    tail = re.sub(
        r"\((?:[^)]*(?:multi|subs|audio|weekly|batch|season|\|)[^)]*)\)",
        " ",
        tail,
        flags=re.I,
    )

    stop_match = _EPISODE_TITLE_STOP_REGEX.search(tail)
    if stop_match:
        tail = tail[: stop_match.start()]

    cleaned = re.sub(r"\s{2,}", " ", tail).strip(" -:|/._")
    if not cleaned or not re.search(r"[A-Za-z]", cleaned):
        return None

    return cleaned


def _extract_audio_langs(title_lower):
    audio_langs = []

    if "english dub" in title_lower or "eng dub" in title_lower:
        audio_langs.append("English")

    if "german dub" in title_lower or "ger dub" in title_lower:
        audio_langs.append("German")

    return audio_langs


def _extract_resolution(title_lower):
    if "2160p" in title_lower or "4k" in title_lower:
        return "2160p"
    if "1080p" in title_lower:
        return "1080p"
    if "720p" in title_lower:
        return "720p"
    if "480p" in title_lower:
        return "480p"
    return ""


def _extract_audio_codec(raw_title):
    for pattern, resolver in _AUDIO_CODEC_PATTERNS:
        match = pattern.search(raw_title)
        if match:
            return resolver(match)

    return ""


def _extract_video_codec(raw_title):
    if re.search(r"\bAV1\b", raw_title, re.I):
        return "AV1"
    if re.search(r"\bx265\b", raw_title, re.I):
        return "x265"
    if re.search(r"\bHEVC\b", raw_title, re.I):
        return "HEVC"
    if re.search(r"\bH[\s.]?265\b", raw_title, re.I):
        return "H265"
    if re.search(r"\bAVC\b", raw_title, re.I):
        return "AVC"
    if re.search(r"\bx264\b", raw_title, re.I):
        return "x264"
    if re.search(r"\bH[\s.]?264\b", raw_title, re.I):
        return "H264"
    if re.search(r"\bXVID\b", raw_title, re.I):
        return "Xvid"
    if re.search(r"\bMPEG\b", raw_title, re.I):
        return "MPEG"
    if re.search(r"\bVC-?1\b", raw_title, re.I):
        return "VC1"
    return ""


def _extract_source(title_lower):
    service = ""
    for pattern, label in _SOURCE_TAG_PATTERNS:
        if pattern.search(title_lower):
            service = label
            break

    source = ""
    if "webrip" in title_lower:
        source = "WEBRip"
    elif "web-dl" in title_lower or "webdl" in title_lower:
        source = "WEB-DL"
    elif re.search(r"\bweb\b", title_lower):
        source = "WEB"
    elif (
        "blu-ray" in title_lower
        or "bluray" in title_lower
        or re.search(r"\bbd\b", title_lower)
    ):
        source = "BluRay"
    elif "hdtv" in title_lower or "tvrip" in title_lower:
        source = "HDTV"

    if service and source:
        return f"{service}.{source}"
    if source:
        return source
    return service


def _extract_release_group(raw_title):
    stripped = re.sub(r"\s*\([^)]*\)\s*$", "", raw_title).strip()

    leading_match = re.match(r"^\[([^\]]+)\]", stripped)
    if leading_match:
        return _normalize_group_name(leading_match.group(1))

    trailing_match = re.search(r"-([A-Za-z0-9][A-Za-z0-9._-]*)$", stripped)
    if trailing_match:
        return _normalize_group_name(trailing_match.group(1))

    return ""


def _normalize_group_name(group_name):
    return re.sub(r"[^A-Za-z0-9]+", "", group_name or "")


def _extract_season_part(raw_title):
    match = _PART_REGEX.search(raw_title)
    if match:
        return int(match.group(1))
    return None


def _parse_listing_datetime(date_text):
    date_text = str(date_text or "").replace("Date/time submitted:", "").strip()
    if not date_text:
        return ""

    now = datetime.utcnow()
    lowered = date_text.lower()

    if lowered.startswith("today "):
        parsed = datetime.strptime(date_text[6:], "%H:%M")
        parsed = now.replace(
            hour=parsed.hour,
            minute=parsed.minute,
            second=0,
            microsecond=0,
        )
    elif lowered.startswith("yesterday "):
        parsed = datetime.strptime(date_text[10:], "%H:%M")
        parsed = (now - timedelta(days=1)).replace(
            hour=parsed.hour,
            minute=parsed.minute,
            second=0,
            microsecond=0,
        )
    else:
        parsed = datetime.strptime(date_text, "%d/%m/%Y %H:%M")

    return parsed.strftime("%a, %d %b %Y %H:%M:%S +0000")


def _resolve_title_context(series_title, raw_title, release_info):
    raw_base_title = _extract_series_title_from_raw_title(raw_title)
    page_title = _clean_series_title(series_title) or raw_base_title or raw_title

    if (
        release_info.season is None
        and release_info.episode_min is None
        and raw_base_title
        and len(raw_base_title) > len(page_title)
    ):
        page_title = raw_base_title

    season_names = _get_cached_season_names(page_title)

    if not season_names:
        trailing_title, trailing_season = _resolve_trailing_season_context(page_title)
        if trailing_season is not None and (
            release_info.season is None or int(release_info.season) == trailing_season
        ):
            release_info.season = trailing_season
            return trailing_title, release_info

        if release_info.season is None:
            release_info.season = _extract_season_hint_from_title(page_title)
        return page_title, release_info

    canonical_title = season_names.get("all") or page_title
    if isinstance(canonical_title, list):
        canonical_title = canonical_title[0] if canonical_title else page_title

    resolved_season = release_info.season
    if resolved_season is None:
        resolved_season = _extract_season_hint_from_title(page_title)

    if resolved_season is None and release_info.episode_min is None:
        resolved_season = _match_xem_season(page_title, season_names)

    release_info.season = resolved_season
    if release_info.season is None:
        return page_title, release_info

    return str(canonical_title or page_title), release_info


def _match_xem_season(page_title, season_names):
    sanitized_page_title = sanitize_string(page_title)
    exact_match = None
    partial_matches = []

    for season_key, names_by_lang in season_names.items():
        if not str(season_key).isdigit():
            continue
        for names in names_by_lang.values():
            for name in names:
                sanitized_name = sanitize_string(name)
                if not sanitized_name:
                    continue
                if sanitized_page_title == sanitized_name:
                    exact_match = int(season_key)
                    break
                if (
                    sanitized_page_title in sanitized_name
                    or sanitized_name in sanitized_page_title
                ):
                    partial_matches.append((len(sanitized_name), int(season_key)))
            if exact_match is not None:
                break
        if exact_match is not None:
            break

    if exact_match is not None:
        return exact_match

    if partial_matches:
        partial_matches.sort(reverse=True)
        return partial_matches[0][1]

    return None


@lru_cache(maxsize=256)
def _get_cached_season_names(page_title):
    try:
        return get_all_season_names(page_title)
    except Exception:
        return None



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import base64
import json
import re
import time
from typing import List, Optional
from urllib.parse import urlparse

from bs4 import BeautifulSoup

from quasarr.constants import DOWNLOAD_REQUEST_TIMEOUT_SECONDS
from quasarr.downloads.linkcrypters.al import decrypt_content, solve_captcha
from quasarr.downloads.sources.helpers.abstract_source import AbstractDownloadSource
from quasarr.downloads.sources.helpers.anime_title import (
    ReleaseInfo,
)
from quasarr.downloads.sources.helpers.anime_title import (
    guess_release_title as _shared_guess_release_title,
)
from quasarr.downloads.sources.helpers.anime_title import (
    inject_subtitle_tokens_in_title as _shared_inject_subtitle_tokens_in_title,
)
from quasarr.downloads.sources.helpers.anime_title import (
    subtitle_lang_to_alpha2 as _shared_subtitle_lang_to_alpha2,
)
from quasarr.downloads.sources.helpers.anime_title import (
    subtitle_tokens as _shared_subtitle_tokens,
)
from quasarr.providers.hostname_issues import mark_hostname_issue
from quasarr.providers.log import debug, info, trace
from quasarr.providers.sessions.al import (
    fetch_via_flaresolverr,
    fetch_via_requests_session,
    invalidate_session,
    retrieve_and_validate_session,
    unwrap_flaresolverr_body,
)
from quasarr.providers.statistics import StatsHelper
from quasarr.providers.utils import is_flaresolverr_available


class Source(AbstractDownloadSource):
    initials = "al"

    def get_download_links(self, shared_state, url, mirrors, title, password):
        """
        AL source handler. Returns plain download links automatically by solving CAPTCHA.

        Note: The 'password' parameter is intentionally repurposed as release_id
        to ensure we download the correct release from the search results.
        This is set by the search module, not a user password.
        """
        # Check if FlareSolverr is available - AL requires it
        if not is_flaresolverr_available(shared_state):
            info(
                "This source requires FlareSolverr which is not configured. "
                "Please configure FlareSolverr in the web UI to use this site."
            )
            return {}

        release_id = _normalize_release_id(password)

        al = shared_state.values["config"]("Hostnames").get(Source.initials)

        sess = retrieve_and_validate_session(shared_state)
        if not sess:
            info(f"Could not retrieve valid session for {al}")
            mark_hostname_issue(Source.initials, "download", "Session error")
            return {}

        details_page = fetch_via_flaresolverr(
            shared_state,
            "GET",
            url,
            timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
        )
        details_html = details_page.get("text", "")
        if not details_html:
            info(f"Failed to load details page for {title} at {url}")
            return {}

        episode_in_title = _extract_episode(title)
        if episode_in_title:
            selection = episode_in_title - 1  # Convert to zero-based index
        else:
            selection = "cnl"

        title, release_id = _check_release(
            shared_state, details_html, release_id, title, episode_in_title
        )
        if release_id == 0:
            info(f"No valid release ID found for {title} - Download failed!")
            return {}

        anime_identifier = url.rstrip("/").split("/")[-1]

        info(f'Selected "Release {release_id}" from {url}')

        links = []
        try:
            raw_request = json.dumps(
                ["media", anime_identifier, "downloads", release_id, selection]
            )
            b64 = base64.b64encode(raw_request.encode("ascii")).decode("ascii")

            post_url = f"https://www.{al}/ajax/captcha"
            payload = {"enc": b64, "response": "nocaptcha"}

            result = fetch_via_flaresolverr(
                shared_state,
                method="POST",
                target_url=post_url,
                post_data=payload,
                timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
            )

            status = result.get("status_code")
            if not status == 200:
                info(f"FlareSolverr returned HTTP {status} for captcha request")
                StatsHelper(shared_state).increment_failed_decryptions_automatic()
                return {}
            else:
                text = result.get("text", "")
                try:
                    response_json = result["json"]
                except ValueError:
                    info(f"Unexpected response when initiating captcha: {text}")
                    StatsHelper(shared_state).increment_failed_decryptions_automatic()
                    return {}

                code = response_json.get("code", "")
                message = response_json.get("message", "")
                content_items = response_json.get("content", [])

                tries = 0
                if code == "success" and content_items:
                    info("CAPTCHA not required")
                elif message == "cnl_login":
                    info("Login expired, re-creating session...")
                    invalidate_session(shared_state)
                else:
                    tries = 0
                    while tries < 3:
                        try:
                            tries += 1
                            info(
                                f"Starting attempt {tries} to solve CAPTCHA for "
                                f"{f'episode {episode_in_title}' if selection and selection != 'cnl' else 'all links'}"
                            )
                            attempt = solve_captcha(
                                Source.initials,
                                shared_state,
                                fetch_via_flaresolverr,
                                fetch_via_requests_session,
                            )

                            solved = (
                                unwrap_flaresolverr_body(attempt.get("response")) == "1"
                            )
                            captcha_id = attempt.get("captcha_id", None)

                            if solved and captcha_id:
                                payload = {
                                    "enc": b64,
                                    "response": "captcha",
                                    "captcha-idhf": 0,
                                    "captcha-hf": captcha_id,
                                }
                                check_solution = fetch_via_flaresolverr(
                                    shared_state,
                                    method="POST",
                                    target_url=post_url,
                                    post_data=payload,
                                    timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
                                )
                                try:
                                    response_json = check_solution.get("json", {})
                                except ValueError as e:
                                    raise RuntimeError(
                                        f"Unexpected /ajax/captcha response: {check_solution.get('text', '')}"
                                    ) from e

                                code = response_json.get("code", "")
                                message = response_json.get("message", "")
                                content_items = response_json.get("content", [])

                                if code == "success":
                                    if content_items:
                                        info(
                                            "CAPTCHA solved successfully on attempt {}.".format(
                                                tries
                                            )
                                        )
                                        break
                                    else:
                                        info(
                                            "CAPTCHA was solved, but no links are available for the selection!"
                                        )
                                        StatsHelper(
                                            shared_state
                                        ).increment_failed_decryptions_automatic()
                                        return {}
                                elif message == "cnl_login":
                                    info("Login expired, re-creating session...")
                                    invalidate_session(shared_state)
                                else:
                                    info(
                                        f"CAPTCHA POST returned code={code}, message={message}. Retrying... (attempt {tries})"
                                    )

                                    if "slowndown" in str(message).lower():
                                        wait_period = 30
                                        info(
                                            f"CAPTCHAs solved too quickly. Waiting {wait_period} seconds before next attempt..."
                                        )
                                        time.sleep(wait_period)
                            else:
                                info(
                                    f"CAPTCHA solver returned invalid solution, retrying... (attempt {tries})"
                                )

                        except RuntimeError as e:
                            info(f"Error solving CAPTCHA: {e}")
                            mark_hostname_issue(
                                Source.initials,
                                "download",
                                str(e) if "e" in dir() else "Download error",
                            )
                        else:
                            info(
                                f"CAPTCHA solver returned invalid solution, retrying... (attempt {tries})"
                            )

                if code != "success":
                    info(
                        f"CAPTCHA solution failed after {tries} attempts. Your IP is likely banned - "
                        f"Code: {code}, Message: {message}"
                    )
                    invalidate_session(shared_state)
                    StatsHelper(shared_state).increment_failed_decryptions_automatic()
                    return {}

                try:
                    links = decrypt_content(content_items, mirrors)
                    debug(f"Decrypted URLs: {links}")
                except Exception as e:
                    info(f"Error during decryption: {e}")
                    mark_hostname_issue(
                        Source.initials,
                        "download",
                        str(e) if "e" in dir() else "Download error",
                    )
        except Exception as e:
            info(f"Error loading download: {e}")
            mark_hostname_issue(
                Source.initials,
                "download",
                str(e) if "e" in dir() else "Download error",
            )
            invalidate_session(shared_state)

        success = bool(links)
        if success:
            StatsHelper(shared_state).increment_captcha_decryptions_automatic()
        else:
            StatsHelper(shared_state).increment_failed_decryptions_automatic()

        links_with_mirrors = [[url, _derive_mirror(url)] for url in links]

        return {"links": links_with_mirrors, "password": f"www.{al}", "title": title}


def _roman_to_int(r: str) -> int:
    roman_map = {"I": 1, "V": 5, "X": 10}
    total = 0
    prev = 0
    for ch in r.upper()[::-1]:
        val = roman_map.get(ch, 0)
        if val < prev:
            total -= val
        else:
            total += val
        prev = val
    return total


def _normalize_release_id(raw_release_id) -> int:
    """
    AL uses download-link password to transport release IDs.
    Older/invalid payloads can provide None/empty/non-numeric values.
    """
    if raw_release_id in (None, "", "None"):
        return 0
    try:
        return int(str(raw_release_id).strip())
    except (TypeError, ValueError):
        return 0


def _derive_mirror(url):
    try:
        hostname = urlparse(url).netloc.lower()
        if hostname.startswith("www."):
            hostname = hostname[4:]
        parts = hostname.split(".")
        return parts[-2] if len(parts) >= 2 else hostname
    except:
        return "unknown"


def _extract_season_from_synonyms(soup):
    """
    Returns the first season found as "Season N" in the Synonym(s) <td>, or None.
    Only scans the synonyms cell—no fallback to whole document.
    """
    syn_td = None
    for tr in soup.select("tr"):
        th = tr.find("th")
        if th and "synonym" in th.get_text(strip=True).lower():
            syn_td = tr.find("td")
            break

    if not syn_td:
        return None

    text = syn_td.get_text(" ", strip=True)

    synonym_season_patterns = [
        re.compile(r"\b(?:Season|Staffel)\s*0?(\d+)\b", re.IGNORECASE),
        re.compile(r"\b0?(\d+)(?:st|nd|rd|th)\s+Season\b", re.IGNORECASE),
        re.compile(r"\b(\d+)\.\s*Staffel\b", re.IGNORECASE),
        re.compile(r"\bS0?(\d+)\b", re.IGNORECASE),  # S02, s2, etc.
        re.compile(r"\b([IVXLCDM]+)\b(?=\s*$)"),  # uppercase Roman at end
    ]

    for pat in synonym_season_patterns:
        m = pat.search(text)
        if not m:
            continue

        tok = m.group(0)
        # Digit match → extract number
        dm = re.search(r"(\d+)", tok)
        if dm:
            return int(dm.group(1))
        # Uppercase Roman → convert & return
        if tok.isupper() and re.fullmatch(r"[IVXLCDM]+", tok):
            return _roman_to_int(tok)

    return None


def _find_season_in_release_notes(soup):
    """
    Iterates through all <tr> rows with a "Release Notes" <th> (case-insensitive).
    Returns the first season number found as an int, or None if not found.
    """
    patterns = [
        re.compile(r"\b(?:Season|Staffel)\s*0?(\d+)\b", re.IGNORECASE),
        re.compile(r"\b0?(\d+)(?:st|nd|rd|th)\s+Season\b", re.IGNORECASE),
        re.compile(r"\b(\d+)\.\s*Staffel\b", re.IGNORECASE),
        re.compile(r"\bS(?:eason)?0?(\d+)\b", re.IGNORECASE),
        re.compile(r"\b([IVXLCDM]+)\b(?=\s*$)"),  # uppercase Roman at end
    ]

    for tr in soup.select("tr"):
        th = tr.find("th")
        if not th:
            continue

        header = th.get_text(strip=True)
        if "release " not in header.lower():  # release notes or release anmerkungen
            continue

        td = tr.find("td")
        if not td:
            continue

        content = td.get_text(" ", strip=True)
        for pat in patterns:
            m = pat.search(content)
            if not m:
                continue

            token = m.group(1)
            # Roman numeral detection only uppercase
            if pat.pattern.endswith("(?=\\s*$)"):
                if token.isupper():
                    return _roman_to_int(token)
                else:
                    continue
            return int(token)

    return None


def _extract_season_number_from_title(page_title, release_type, release_title=""):
    """
    Extracts the season number from the given page title.

    Priority is given to standard patterns like S01/E01 or R2 in the optional release title.
    If no match is found, it attempts to extract based on keywords like "Season"/"Staffel"
    or trailing numbers/roman numerals in the page title.

    Args:
        page_title (str): The title of the page, used as a fallback.
        release_type (str): The type of release (e.g., 'series').
        release_title (Optional, str): The title of the release.

    Returns:
        int: The extracted or inferred season number. Defaults to 1 if not found.
    """
    season_num = None

    if release_title:
        match = re.search(
            r"\.(?:S(\d{1,4})|R(2))(?:E\d{1,4})?", release_title, re.IGNORECASE
        )
        if match:
            if match.group(1) is not None:
                season_num = int(match.group(1))
            elif match.group(2) is not None:
                season_num = int(match.group(2))

    if season_num is None:
        page_title = page_title or ""
        if (
            "staffel" in page_title.lower()
            or "season" in page_title.lower()
            or release_type == "series"
        ):
            match = re.search(
                r"\b(?:Season|Staffel)\s+(\d+|[IVX]+)\b|\bR(2)\b",
                page_title,
                re.IGNORECASE,
            )
            if match:
                if match.group(1) is not None:
                    num = match.group(1)
                    season_num = int(num) if num.isdigit() else _roman_to_int(num)
                elif match.group(2) is not None:
                    season_num = int(match.group(2))
            else:
                trailing_match = re.search(
                    r"\s+([2-9]\d*|[IVXLCDM]+)\s*$", page_title, re.IGNORECASE
                )
                if trailing_match:
                    num = trailing_match.group(1)
                    season_candidate = int(num) if num.isdigit() else _roman_to_int(num)
                    if season_candidate >= 2:
                        season_num = season_candidate

            if season_num is None:
                season_num = 1

    return season_num


def _subtitle_lang_to_alpha2(lang: str) -> Optional[str]:
    return _shared_subtitle_lang_to_alpha2(lang)


def _subtitle_tokens(subtitle_langs: List[str]) -> List[str]:
    return _shared_subtitle_tokens(subtitle_langs)


def _inject_subtitle_tokens_in_title(title: str, subtitle_langs: List[str]) -> str:
    return _shared_inject_subtitle_tokens_in_title(title, subtitle_langs)


def _parse_info_from_feed_entry(block, series_page_title, release_type) -> ReleaseInfo:
    """
    Parse a BeautifulSoup block from the feed entry into ReleaseInfo.
    """
    text = block.get_text(separator=" ", strip=True)

    # detect season
    season_num = _extract_season_number_from_title(series_page_title, release_type)

    # detect episodes
    episode_min: Optional[int] = None
    episode_max: Optional[int] = None
    m_ep = re.search(r"Episode\s+(\d+)(?:-(\d+))?", text)
    if m_ep:
        episode_min = int(m_ep.group(1))
        episode_max = int(m_ep.group(2)) if m_ep.group(2) else episode_min

    # parse audio flags
    audio_langs: List[str] = []
    audio_icon = block.find("i", class_="fa-volume-up")
    if audio_icon:
        for sib in audio_icon.find_next_siblings():
            if sib.name == "i" and "fa-closed-captioning" in sib.get("class", []):
                break
            if sib.name == "i" and "flag" in sib.get("class", []):
                code = sib["class"][1].replace("flag-", "").lower()
                audio_langs.append(
                    {"jp": "Japanese", "de": "German", "en": "English"}.get(
                        code, code.title()
                    )
                )

    # parse subtitle flags
    subtitle_langs: List[str] = []
    subtitle_icon = block.find("i", class_="fa-closed-captioning")
    if subtitle_icon:
        for sib in subtitle_icon.find_next_siblings():
            if sib.name == "i" and "flag" in sib.get("class", []):
                code = sib["class"][1].replace("flag-", "").lower()
                subtitle_langs.append(
                    {"jp": "Japanese", "de": "German", "en": "English"}.get(
                        code, code.title()
                    )
                )

    # resolution
    m_res = re.search(r":\s*([0-9]{3,4}p)", text, re.IGNORECASE)
    resolution = m_res.group(1) if m_res else "1080p"

    # source not available in feed
    source = "WEB-DL"
    # video codec not available in feed
    video = "x264"

    # release group
    span = block.find("span")
    if span:
        grp = span.get_text().split(":", 1)[-1].strip()
        release_group = grp.replace(" ", "").replace("-", "")
    else:
        release_group = ""

    return ReleaseInfo(
        release_title=None,
        audio_langs=audio_langs,
        subtitle_langs=subtitle_langs,
        episode_title=None,
        resolution=resolution,
        audio="",
        video=video,
        source=source,
        release_group=release_group,
        season_part=None,
        season=season_num,
        episode_min=episode_min,
        episode_max=episode_max,
    )


def _parse_info_from_download_item(
    tab,
    content,
    page_title=None,
    release_type=None,
    requested_season=False,
    requested_episode=None,
) -> ReleaseInfo:
    """
    Parse a BeautifulSoup 'tab' from a download item into ReleaseInfo.
    """
    # notes
    notes_td = tab.select_one("tr:has(th>i.fa-info) td")
    notes_text = notes_td.get_text(strip=True) if notes_td else ""
    notes_lower = notes_text.lower()
    trace(
        f"parse_download_item tab={tab.get('id', '')} "
        f"page_title='{page_title}' requested_season={requested_season} "
        f"requested_episode={requested_episode} notes_text='{notes_text}'"
    )

    release_title = None
    if notes_text:
        rn_with_dots = notes_text.replace(" ", ".").replace(".-.", "-")
        rn_no_dot_duplicates = re.sub(r"\.{2,}", ".", rn_with_dots)
        trace(
            f"notes_transformed tab={tab.get('id', '')} "
            f"dotted='{rn_with_dots}' deduped='{rn_no_dot_duplicates}'"
        )
        if "." in rn_with_dots and "-" in rn_with_dots:
            # Check if string ends with Group tag (word after dash) - this should prevent false positives
            if re.search(r"-[\s.]?\w+$", rn_with_dots):
                release_title = rn_no_dot_duplicates
                trace(
                    f"release_title accepted from notes "
                    f"tab={tab.get('id', '')}: '{release_title}'"
                )

    # resolution
    res_td = tab.select_one("tr:has(th>i.fa-desktop) td")
    resolution = "1080p"
    if res_td:
        match = re.search(r"(\d+)\s*x\s*(\d+)", res_td.get_text(strip=True))
        if match:
            h = int(match.group(2))
            resolution = "2160p" if h >= 2000 else "1080p" if h >= 1000 else "720p"

    # audio and subtitles
    audio_codes = [
        icon["class"][1].replace("flag-", "")
        for icon in tab.select("tr:has(th>i.fa-volume-up) i.flag")
    ]
    audio_langs = [
        {"jp": "Japanese", "de": "German", "en": "English"}.get(c, c.title())
        for c in audio_codes
    ]
    sub_codes = [
        icon["class"][1].replace("flag-", "")
        for icon in tab.select("tr:has(th>i.fa-closed-captioning) i.flag")
    ]
    subtitle_langs = [
        {"jp": "Japanese", "de": "German", "en": "English"}.get(c, c.title())
        for c in sub_codes
    ]

    # audio codec
    if "flac" in notes_lower:
        audio = "FLAC"
    elif "aac" in notes_lower:
        audio = "AAC"
    elif "opus" in notes_lower:
        audio = "Opus"
    elif "mp3" in notes_lower:
        audio = "MP3"
    elif "pcm" in notes_lower:
        audio = "PCM"
    elif "dts" in notes_lower:
        audio = "DTS"
    elif "ac3" in notes_lower or "eac3" in notes_lower:
        audio = "AC3"
    else:
        audio = ""

    # source
    if re.search(r"(web-dl|webdl|webrip)", notes_lower):
        source = "WEB-DL"
    elif re.search(r"(blu-ray|\bbd\b|bluray)", notes_lower):
        source = "BluRay"
    elif re.search(r"(hdtv|tvrip)", notes_lower):
        source = "HDTV"
    else:
        source = "WEB-DL"

    if "265" in notes_lower or "hevc" in notes_lower:
        video = "x265"
    elif "av1" in notes_lower:
        video = "AV1"
    elif "avc" in notes_lower:
        video = "AVC"
    elif "xvid" in notes_lower:
        video = "Xvid"
    elif "mpeg" in notes_lower:
        video = "MPEG"
    elif "vc-1" in notes_lower:
        video = "VC-1"
    else:
        video = "x264"

    # release group
    grp_td = tab.select_one("tr:has(th>i.fa-child) td")
    if grp_td:
        grp = grp_td.get_text(strip=True)
        release_group = grp.replace(" ", "").replace("-", "")
    else:
        release_group = ""

    # determine season
    if requested_season:
        season_num = _extract_season_from_synonyms(content)
        if not season_num:
            season_num = _find_season_in_release_notes(content)
        if not season_num:
            season_num = _extract_season_number_from_title(
                page_title, release_type, release_title=release_title
            )
    else:
        season_num = None

    # check if season part info is present
    season_part: Optional[int] = None
    if page_title:
        match = re.search(
            r"(?i)\b(?:Part|Teil)\s+(\d+|[IVX]+)\b", page_title, re.IGNORECASE
        )
        if match:
            num = match.group(1)
            season_part = int(num) if num.isdigit() else _roman_to_int(num)
            part_string = f"Part.{season_part}"
            if release_title and part_string not in release_title:
                release_title = re.sub(
                    r"\.(German|Japanese|English)\.",
                    f".{part_string}.\\1.",
                    release_title,
                    count=1,
                )

    # determine if optional episode exists on release page
    episode_min: Optional[int] = None
    episode_max: Optional[int] = None
    if requested_episode:
        episodes_div = tab.find("div", class_="episodes")
        if episodes_div:
            episode_links = episodes_div.find_all(
                "a", attrs={"data-loop": re.compile(r"^\d+$")}
            )
            total_episodes = len(episode_links)
            if total_episodes > 0:
                ep = int(requested_episode)
                if ep <= total_episodes:
                    episode_min = 1
                    episode_max = total_episodes
                    if release_title:
                        release_title = re.sub(
                            r"(?<=\.)S(\d{1,4})(?=\.)",
                            lambda m: f"S{int(m.group(1)):02d}E{ep:02d}",
                            release_title,
                            count=1,
                            flags=re.IGNORECASE,
                        )

    if release_title:
        release_title = _inject_subtitle_tokens_in_title(release_title, subtitle_langs)

    trace(
        f"parse_download_item result tab={tab.get('id', '')} "
        f"release_title='{release_title}' season={season_num} "
        f"episode_min={episode_min} episode_max={episode_max}"
    )

    return ReleaseInfo(
        release_title=release_title,
        audio_langs=audio_langs,
        subtitle_langs=subtitle_langs,
        episode_title=None,
        resolution=resolution,
        audio=audio,
        video=video,
        source=source,
        release_group=release_group,
        season_part=season_part,
        season=season_num,
        episode_min=episode_min,
        episode_max=episode_max,
    )


def _guess_title(page_title, release_info: ReleaseInfo) -> str:
    return _shared_guess_release_title(page_title, release_info)


def _check_release(shared_state, details_html, release_id, title, episode_in_title):
    soup = BeautifulSoup(details_html, "html.parser")
    release_id = _normalize_release_id(release_id)

    if release_id == 0:
        info(
            "Feed download detected, hard-coding release_id to 1 to achieve successful download"
        )
        release_id = 1
        # The following logic works, but the highest release ID sometimes does not have the desired episode
        #
        # If download was started from the feed, the highest download id is typically the best option
        # panes = soup.find_all("div", class_="tab-pane")
        # max_id = None
        # for pane in panes:
        #     pane_id = pane.get("id", "")
        #     match = re.match(r"download_(\d+)$", pane_id)
        #     if match:
        #         num = int(match.group(1))
        #         if max_id is None or num > max_id:
        #             max_id = num
        # if max_id:
        #     release_id = max_id

    tab = soup.find("div", class_="tab-pane", id=f"download_{release_id}")
    if tab:
        try:
            # We re-guess the title from the details page
            # This ensures, that downloads initiated by the feed (which has limited/incomplete data) yield
            # the best possible title for the download (including resolution, audio, video, etc.)
            page_title_info = soup.find("title").text.strip().rpartition(" (")
            page_title = page_title_info[0].strip()
            release_type_info = page_title_info[2].strip()
            if "serie" in release_type_info.lower():
                release_type = "series"
            else:
                release_type = "movie"

            release_info = _parse_info_from_download_item(
                tab,
                soup,
                page_title=page_title,
                release_type=release_type,
                requested_season=True if release_type == "series" else False,
                requested_episode=episode_in_title,
            )
            real_title = release_info.release_title
            if real_title:
                if real_title.lower() != title.lower():
                    info(
                        f'Identified true release title "{real_title}" on details page'
                    )
                    return real_title, release_id
            else:
                # Overwrite values so guessing the title only applies the requested episode
                if episode_in_title:
                    release_info.episode_min = int(episode_in_title)
                    release_info.episode_max = int(episode_in_title)

                guessed_title = _guess_title(page_title, release_info)
                if guessed_title and guessed_title.lower() != title.lower():
                    info(
                        f'Adjusted guessed release title to "{guessed_title}" from details page'
                    )
                    return guessed_title, release_id
        except Exception as e:
            info(f"Error guessing release title from release: {e}")
            mark_hostname_issue(
                Source.initials,
                "download",
                str(e) if "e" in dir() else "Download error",
            )

    return title, release_id


def _extract_episode(title: str) -> int | None:
    match = re.search(r"\bS\d{1,4}E(\d+)\b(?![\-E\d])", title)
    if match:
        return int(match.group(1))

    if not re.search(r"\bS\d{1,4}\b", title):
        match = re.search(r"\.E(\d+)\b(?![\-E\d])", title)
        if match:
            return int(match.group(1))

    return None



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from urllib.parse import quote_plus, urljoin

import requests
from bs4 import BeautifulSoup

from quasarr.constants import (
    FEED_REQUEST_TIMEOUT_SECONDS,
    SEARCH_CAT_MOVIES,
    SEARCH_CAT_SHOWS,
    SEARCH_CAT_SHOWS_ANIME,
    SEARCH_REQUEST_TIMEOUT_SECONDS,
)
from quasarr.downloads.sources.at import (
    _build_release_info_from_title,
    _extract_direct_links_from_anchors,
    _extract_season_hint_from_title,
    _extract_series_title_from_listing_entry,
    _extract_series_title_from_raw_title,
    _extract_subtitle_langs_from_links,
    _parse_listing_datetime,
    _resolve_title_context,
)
from quasarr.downloads.sources.helpers.anime_title import guess_release_title
from quasarr.providers import shared_state
from quasarr.providers.hostname_issues import clear_hostname_issue, mark_hostname_issue
from quasarr.providers.imdb_metadata import get_localized_title, get_year
from quasarr.providers.log import debug, info, trace, warn
from quasarr.providers.utils import (
    convert_to_mb,
    generate_download_link,
    get_base_search_category_id,
    is_imdb_id,
    match_in_title,
    search_string_in_sanitized_title,
)
from quasarr.providers.xem_metadata import get_season_name
from quasarr.search.sources.helpers.search_release import SearchRelease
from quasarr.search.sources.helpers.search_source import AbstractSearchSource

_SEASON_DASH_EP_REGEX = re.compile(
    r"\bS(?:eason)?\s*0*(\d{1,3})\s*-\s*0*(\d{1,4})(?:\s*[-+]\s*0*(\d{1,4}))?",
    re.I,
)
_SEASON_EP_REGEX = re.compile(
    r"\bS(?:eason)?\s*0*(\d{1,3})\s*[.\-_ ]?(?:E|EP|Episode)\s*0*(\d{1,4})(?:\s*[-+]\s*(?:E|EP|Episode)?\s*0*(\d{1,4}))?\b",
    re.I,
)
_ABSOLUTE_EP_REGEX = re.compile(
    r"\b(?:EP?|Episode)\s*0*(\d{1,4})(?:\s*[-+]\s*0*(\d{1,4}))?\b", re.I
)
_ABSOLUTE_RANGE_REGEX = re.compile(r"[\[(]0*(\d{1,4})\s*[-+]\s*0*(\d{1,4})[\])]", re.I)
_ABSOLUTE_DASH_REGEX = re.compile(
    r"\s-\s0*(\d{1,3})(?:\s*[-+]\s*0*(\d{1,3}))?(?=[\s[(]|$)", re.I
)
_PLAIN_EP_REGEX = re.compile(r"(?:\.|^)[eE](\d{1,4})(?:-(\d{1,4}))?(?=[\.-]|$)")
_SEASON_WORD_REGEX = re.compile(r"\b(?:Season|Staffel)\s*0*(\d{1,3})\b", re.I)
_SIZE_REGEX = re.compile(r"(\d+(?:\.\d+)?)\s*(KB|MB|GB|TB)", re.I)


class Source(AbstractSearchSource):
    initials = "at"
    supports_imdb = True
    supports_phrase = True
    supports_absolute_numbering = True
    supported_categories = [SEARCH_CAT_MOVIES, SEARCH_CAT_SHOWS, SEARCH_CAT_SHOWS_ANIME]

    def feed(
        self, shared_state: shared_state, start_time: float, search_category: str
    ) -> list[SearchRelease]:
        releases = []
        host = shared_state.values["config"]("Hostnames").get(self.initials)
        if not host:
            return releases

        try:
            entries = _load_entries(
                shared_state,
                f"https://{host}/",
                f"https://{host}/?disp=attachments",
                FEED_REQUEST_TIMEOUT_SECONDS,
            )
        except Exception as e:
            warn(f"Error loading feed: {e}")
            mark_hostname_issue(self.initials, "feed", str(e) or "Error occurred")
            return releases

        seen_sources = set()
        for entry in entries:
            title = entry["title"]
            if entry["source"] in seen_sources:
                continue
            if not _matches_category(title, search_category):
                continue

            seen_sources.add(entry["source"])
            releases.append(
                {
                    "details": {
                        "title": title,
                        "hostname": self.initials,
                        "imdb_id": None,
                        "link": generate_download_link(
                            shared_state,
                            title,
                            entry["source"],
                            entry["mb"],
                            "",
                            None,
                            self.initials,
                        ),
                        "size": entry["mb"] * 1024 * 1024,
                        "date": entry["published"],
                        "source": entry["source"],
                    },
                    "type": "protected",
                }
            )

        elapsed = time.time() - start_time
        debug(f"Time taken: {elapsed:.2f}s")

        if releases:
            clear_hostname_issue(self.initials)
        return releases

    def search(
        self,
        shared_state: shared_state,
        start_time: float,
        search_category: str,
        search_string: str = "",
        season: int = None,
        episode: int = None,
    ) -> list[SearchRelease]:
        releases = []
        host = shared_state.values["config"]("Hostnames").get(self.initials)
        if not host:
            return releases

        season = _normalize_search_number(season)
        episode = _normalize_search_number(episode)

        base_category = get_base_search_category_id(search_category)
        if base_category not in (
            SEARCH_CAT_MOVIES,
            SEARCH_CAT_SHOWS,
            SEARCH_CAT_SHOWS_ANIME,
        ):
            warn(f"Unknown search category: {search_category}")
            return releases

        imdb_id = is_imdb_id(search_string)
        if imdb_id:
            localized_title = get_localized_title(shared_state, imdb_id, "en")
            if not localized_title:
                info(f"Could not extract title from IMDb-ID {imdb_id}")
                return releases
            search_string = localized_title

        for variant in _build_search_variants(search_string, imdb_id, season, episode):
            try:
                query = quote_plus(variant["query"])
                entries = _load_entries(
                    shared_state,
                    f"https://{host}/search?q={query}",
                    f"https://{host}/search?q={query}&disp=attachments",
                    SEARCH_REQUEST_TIMEOUT_SECONDS,
                )
            except Exception as e:
                warn(f"Error loading search for {variant['query']}: {e}")
                mark_hostname_issue(self.initials, "search", str(e) or "Error occurred")
                continue

            seen_sources = set()
            for entry in entries:
                title = entry["title"]
                if entry["source"] in seen_sources:
                    continue
                if not _matches_entry_search(variant["match"], entry):
                    continue
                if not _matches_requested_release(
                    title,
                    search_category,
                    season=season,
                    episode=episode,
                    season_locked=variant["season_locked"],
                ):
                    continue

                seen_sources.add(entry["source"])
                releases.append(
                    {
                        "details": {
                            "title": title,
                            "hostname": self.initials,
                            "imdb_id": imdb_id,
                            "link": generate_download_link(
                                shared_state,
                                title,
                                entry["source"],
                                entry["mb"],
                                "",
                                imdb_id,
                                self.initials,
                            ),
                            "size": entry["mb"] * 1024 * 1024,
                            "date": entry["published"],
                            "source": entry["source"],
                        },
                        "type": "protected",
                    }
                )

            if releases:
                break

        elapsed = time.time() - start_time
        debug(f"Time taken: {elapsed:.2f}s")

        if releases:
            clear_hostname_issue(self.initials)
        return releases


def _load_entries(shared_state, listing_url, attachments_url, request_timeout):
    headers = {"User-Agent": shared_state.values["user_agent"]}

    def fetch(url):
        response = requests.get(url, headers=headers, timeout=request_timeout)
        response.raise_for_status()
        return response.text

    responses = {}
    with ThreadPoolExecutor(max_workers=2) as executor:
        futures = {
            executor.submit(fetch, listing_url): "listing",
            executor.submit(fetch, attachments_url): "attachments",
        }
        for future in as_completed(futures):
            kind = futures[future]
            try:
                responses[kind] = future.result()
            except Exception:
                if kind == "listing":
                    raise
                responses[kind] = ""

    entries = _parse_listing_page(responses["listing"], listing_url)
    subtitle_langs_by_source = _parse_attachment_page(
        responses.get("attachments", ""), attachments_url
    )

    for entry in entries:
        entry["subtitle_langs"] = subtitle_langs_by_source.get(entry["source"], [])
        release_info = _build_release_info_from_title(
            entry["raw_title"], entry["subtitle_langs"]
        )
        entry["release_info"] = release_info

    season_hints_by_series = _build_season_hints(entries)

    for entry in entries:
        release_info = entry["release_info"]
        if release_info.season is None:
            season_hint = season_hints_by_series.get(_series_hint_key(entry))
            if season_hint is not None:
                release_info.season = season_hint

        page_title, release_info = _resolve_title_context(
            entry["series_title"]
            or _extract_series_title_from_raw_title(entry["raw_title"])
            or entry["raw_title"],
            entry["raw_title"],
            release_info,
        )
        entry["title"] = guess_release_title(page_title, release_info)

    return entries


def _parse_listing_page(html_text, page_url):
    entries = []
    soup = BeautifulSoup(html_text, "html.parser")

    for item in soup.select("div.home_list_entry"):
        title_link = item.select_one("div.link a[href]")
        links_div = item.select_one("div.links")
        size_node = item.select_one("div.size")
        if not title_link or not links_div:
            continue

        if not _extract_direct_links_from_anchors(links_div.find_all("a", href=True)):
            continue

        source = urljoin(page_url, title_link.get("href", "").strip())
        if not source:
            continue

        entries.append(
            {
                "raw_title": title_link.get_text(" ", strip=True),
                "series_title": _extract_series_title_from_listing_entry(item),
                "source": source,
                "published": _extract_listing_date(item),
                "mb": _extract_listing_size_mb(size_node),
                "subtitle_langs": [],
            }
        )

    return entries


def _parse_attachment_page(html_text, page_url):
    subtitle_langs_by_source = {}
    if not html_text:
        return subtitle_langs_by_source

    soup = BeautifulSoup(html_text, "html.parser")
    for entry in soup.select("div.home_list_entry"):
        title_link = entry.select_one("div.link a[href]")
        links_div = entry.select_one("div.links")
        if not title_link or not links_div:
            continue

        subtitle_langs = _extract_subtitle_langs_from_links(links_div)
        if subtitle_langs:
            subtitle_langs_by_source[urljoin(page_url, title_link["href"].strip())] = (
                subtitle_langs
            )

    return subtitle_langs_by_source


def _extract_listing_date(item):
    date_node = item.select_one("div.date[title]") or item.select_one(
        "div.date_icon[title]"
    )
    if not date_node:
        return ""

    return _parse_listing_datetime(date_node.get("title"))


def _extract_listing_size_mb(size_node):
    if not size_node:
        return 0

    match = _SIZE_REGEX.search(size_node.get_text(" ", strip=True))
    if not match:
        return 0

    return convert_to_mb({"size": match.group(1), "sizeunit": match.group(2).upper()})


def _build_search_variants(search_string, imdb_id, season, episode):
    variants = []
    season = _normalize_search_number(season)
    episode = _normalize_search_number(episode)

    def append_variants_for_query(query, match, season_locked):
        cleaned_query = str(query or "").strip()
        cleaned_match = str(match or "").strip()
        if not cleaned_query or not cleaned_match:
            return

        variants.append(
            {
                "query": cleaned_query,
                "match": cleaned_match,
                "season_locked": season_locked,
            }
        )

        if episode is None:
            return

        if season is not None:
            variants.append(
                {
                    "query": f"{cleaned_query} S{season:02d}E{episode:02d}",
                    "match": cleaned_match,
                    "season_locked": season_locked,
                }
            )

        variants.append(
            {
                "query": f"{cleaned_query} {episode}",
                "match": cleaned_match,
                "season_locked": season_locked,
            }
        )

    if season is not None:
        append_variants_for_query(
            f"{search_string} Season {season}",
            search_string,
            True,
        )

        xem_name = get_season_name(search_string, season, "en")
        if xem_name:
            append_variants_for_query(xem_name, xem_name, True)

    if imdb_id and season is None:
        year = get_year(imdb_id)
        if year:
            append_variants_for_query(
                f"{search_string} {year}",
                search_string,
                False,
            )

    append_variants_for_query(search_string, search_string, False)

    deduped_variants = []
    seen = set()
    for variant in variants:
        key = variant["query"].lower().strip()
        if not key or key in seen:
            continue
        seen.add(key)
        deduped_variants.append(variant)

    return deduped_variants


def _build_season_hints(entries):
    season_hints_by_series = {}

    for entry in entries:
        release_info = entry["release_info"]
        if release_info.season is None:
            explicit_hint = _extract_season_hint_from_title(
                entry["series_title"]
                or _extract_series_title_from_raw_title(entry["raw_title"])
            )
            if explicit_hint is not None:
                release_info.season = explicit_hint

        if release_info.season is None:
            continue

        key = _series_hint_key(entry)
        season_hints_by_series.setdefault(key, set()).add(release_info.season)

    return {
        key: next(iter(seasons))
        for key, seasons in season_hints_by_series.items()
        if len(seasons) == 1
    }


def _series_hint_key(entry):
    return (
        (
            entry["series_title"]
            or _extract_series_title_from_raw_title(entry["raw_title"])
        )
        .strip()
        .lower()
    )


def _matches_entry_search(search_string, entry):
    for candidate in [entry["title"], entry["raw_title"], entry["series_title"]]:
        if candidate and search_string_in_sanitized_title(search_string, candidate):
            return True

    trace(
        f"Search string '{search_string}' doesn't match '{entry['raw_title']}' or '{entry['title']}'"
    )
    return False


def _matches_category(title, search_category):
    base_category = get_base_search_category_id(search_category)

    if base_category == SEARCH_CAT_MOVIES:
        return not _looks_serial_release(title)

    if base_category in (SEARCH_CAT_SHOWS, SEARCH_CAT_SHOWS_ANIME):
        return _looks_serial_release(title)

    return False


def _matches_requested_release(title, search_category, season, episode, season_locked):
    season = _normalize_search_number(season)
    episode = _normalize_search_number(episode)

    if not _matches_category(title, search_category):
        return False

    if season is None and episode is None:
        return True

    if season is not None and episode is None:
        return _matches_season_pack(title, season)

    if match_in_title(title, season, episode):
        return True

    explicit_seasons = _extract_season_numbers(title)

    if season is not None:
        if _matches_season_dash_episode(title, season, episode):
            return True

        if explicit_seasons and season not in explicit_seasons:
            return False

        if season_locked:
            return _matches_absolute_episode(title, episode)

        return False

    if episode is not None:
        return _matches_absolute_episode(title, episode)

    return False


def _looks_serial_release(title):
    normalized_title = str(title or "")

    if match_in_title(normalized_title):
        return True
    if re.search(r"\bS\d{1,3}E\d{1,4}(?:-\d{1,4})?\b", normalized_title, re.I):
        return True
    if _SEASON_DASH_EP_REGEX.search(normalized_title):
        return True
    if _ABSOLUTE_EP_REGEX.search(normalized_title):
        return True
    if _ABSOLUTE_RANGE_REGEX.search(normalized_title):
        return True
    if _ABSOLUTE_DASH_REGEX.search(normalized_title):
        return True
    if _PLAIN_EP_REGEX.search(normalized_title):
        return True
    if _SEASON_WORD_REGEX.search(normalized_title):
        return True

    return False


def _matches_season_dash_episode(title, season, episode):
    for match in _SEASON_DASH_EP_REGEX.finditer(title):
        if int(match.group(1)) != int(season):
            continue

        if episode is None:
            return True

        start = int(match.group(2))
        end = int(match.group(3) or start)
        if start <= int(episode) <= end:
            return True

    return False


def _matches_absolute_episode(title, episode):
    episode = _normalize_search_number(episode)
    if episode is None:
        return False

    for start, end in _extract_absolute_ranges(title):
        if start <= episode <= end:
            return True

    return False


def _matches_season_pack(title, season):
    explicit_seasons = _extract_season_numbers(title)
    if explicit_seasons and season not in explicit_seasons:
        return False

    episode_ranges = _extract_episode_ranges(title, season)
    if any(start == end for start, end in episode_ranges):
        return False

    normalized_title = str(title or "")
    if explicit_seasons:
        return True

    if season == 1 and re.search(
        r"\b(?:batch|complete|全集|pack)\b", normalized_title, re.I
    ):
        return True

    return False


def _extract_absolute_ranges(title):
    ranges = []

    for pattern in (
        _ABSOLUTE_EP_REGEX,
        _ABSOLUTE_RANGE_REGEX,
        _ABSOLUTE_DASH_REGEX,
        _PLAIN_EP_REGEX,
    ):
        for match in pattern.finditer(title):
            start = int(match.group(1))
            if match.lastindex and match.lastindex >= 2 and match.group(2):
                end = int(match.group(2))
            else:
                end = start
            ranges.append((start, end))

    return ranges


def _extract_episode_ranges(title, season=None):
    ranges = []

    for match in _SEASON_EP_REGEX.finditer(title):
        match_season = int(match.group(1))
        if season is not None and match_season != int(season):
            continue

        start = int(match.group(2))
        end = int(match.group(3) or start)
        ranges.append((start, end))

    if season is None or int(season) == 1:
        ranges.extend(_extract_absolute_ranges(title))

    deduped_ranges = []
    seen = set()
    for start, end in ranges:
        key = (start, end)
        if key in seen:
            continue
        seen.add(key)
        deduped_ranges.append(key)

    return deduped_ranges


def _extract_season_numbers(title):
    seasons = set()

    for match in re.finditer(r"\bS(\d{1,3})(?=E\d{1,4}\b|\b)", title, re.I):
        seasons.add(int(match.group(1)))

    for match in _SEASON_WORD_REGEX.finditer(title):
        seasons.add(int(match.group(1)))

    return seasons


def _normalize_search_number(value):
    if value is None:
        return None

    if isinstance(value, str):
        value = value.strip()
        if not value:
            return None

    try:
        return int(value)
    except (TypeError, ValueError):
        return None



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import concurrent.futures
import re
import time
from urllib.parse import urljoin, urlparse

import requests
from bs4 import BeautifulSoup

from quasarr.constants import DOWNLOAD_REQUEST_TIMEOUT_SECONDS
from quasarr.downloads.sources.helpers.abstract_source import AbstractDownloadSource
from quasarr.providers.hostname_issues import mark_hostname_issue
from quasarr.providers.log import debug, info


class Source(AbstractDownloadSource):
    initials = "by"

    def get_download_links(self, shared_state, url, mirrors, title, password):
        """
        BY source handler - fetches protected download links from BY iframes.
        """
        by = shared_state.values["config"]("Hostnames").get(Source.initials)
        headers = {
            "User-Agent": shared_state.values["user_agent"],
        }

        mirrors_lower = [m.lower() for m in mirrors] if mirrors else []
        links = []

        try:
            r = requests.get(
                url,
                headers=headers,
                timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
            )
            r.raise_for_status()
            soup = BeautifulSoup(r.text, "html.parser")
            frames = [
                iframe.get("src")
                for iframe in soup.find_all("iframe")
                if iframe.get("src")
            ]

            frame_urls = [src for src in frames if f"https://{by}" in src]
            if not frame_urls:
                debug(f"No iframe hosts found on {url} for {title}.")
                return {"links": []}

            async_results = []

            def fetch(url):
                try:
                    rq = requests.get(
                        url,
                        headers=headers,
                        timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
                    )
                    rq.raise_for_status()
                    return rq.text, url
                except Exception as e:
                    info(f"Error fetching iframe URL: {url}")
                    mark_hostname_issue(Source.initials, "download", str(e))
                    return None, url

            with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
                future_to_url = {executor.submit(fetch, url): url for url in frame_urls}
                for future in concurrent.futures.as_completed(future_to_url):
                    content, source = future.result()
                    if content:
                        async_results.append((content, source))

            url_hosters = []
            for content, _source in async_results:
                host_soup = BeautifulSoup(content, "html.parser")
                link = host_soup.find(
                    "a",
                    href=re.compile(
                        r"https?://(?:www\.)?(?:hide\.cx|filecrypt\.(?:cc|co|to))/container/"
                    ),
                )

                # Fallback to the old format
                if not link:
                    link = host_soup.find("a", href=re.compile(r"/go\.php\?"))

                if not link:
                    continue

                href = link["href"]
                link_hostname = link.text.strip().replace(" ", "")
                hostname_lower = link_hostname.lower()

                if mirrors_lower and not any(
                    m in hostname_lower for m in mirrors_lower
                ):
                    debug(
                        f'Skipping link from "{link_hostname}" (not in desired mirrors "{mirrors}")!'
                    )
                    continue

                url_hosters.append((href, link_hostname))

            def _is_protected_or_auto_link(candidate_url):
                lowered = (candidate_url or "").lower()
                return (
                    "filecrypt." in lowered
                    or "hide." in lowered
                    or "tolink." in lowered
                    or "keeplinks." in lowered
                )

            def resolve_redirect(href_hostname):
                href, _hostname = href_hostname
                current_url = href
                visited = set()

                # If iframe already gives protected/auto URL, return directly.
                if _is_protected_or_auto_link(current_url):
                    return current_url

                for _hop in range(8):
                    if current_url in visited:
                        debug(f"BY redirect loop detected for {current_url}")
                        return None
                    visited.add(current_url)

                    if _is_protected_or_auto_link(current_url):
                        return current_url

                    try:
                        # Resolve redirect chain manually so we can capture final
                        # FileCrypt URL without requesting FileCrypt itself.
                        rq = requests.get(
                            current_url,
                            headers=headers,
                            timeout=DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
                            allow_redirects=False,
                        )
                    except Exception as e:
                        debug(f"Error resolving BY redirect for {current_url}: {e}")
                        return None

                    location = (rq.headers.get("Location") or "").strip()
                    if location:
                        next_url = urljoin(current_url, location)
                        if "/404.html" in next_url:
                            debug(f"BY redirect led to 404 page: {next_url}")
                            return None
                        current_url = next_url
                        continue

                    final_url = (rq.url or current_url).strip()
                    if _is_protected_or_auto_link(final_url):
                        return final_url

                    if "/404.html" in final_url:
                        debug(f"BY redirect led to 404 page: {final_url}")
                        return None

                    if rq.status_code >= 400:
                        debug(
                            f"Error resolving BY redirect: HTTP {rq.status_code} at {final_url}"
                        )
                        return None

                    time.sleep(1)
                    return final_url

                debug(f"BY redirect hop limit exceeded for {href}")
                return None

            for pair in url_hosters:
                resolved_url = resolve_redirect(pair)
                link_hostname = pair[1]

                if not resolved_url:
                    continue

                # Protected/auto links must be returned as-is so downstream
                # classification stores them in CAPTCHA flow instead of direct.
                if _is_protected_or_auto_link(resolved_url):
                    links.append([resolved_url, "filecrypt"])
                    continue

                if not link_hostname and resolved_url:
                    link_hostname = urlparse(resolved_url).hostname

                if link_hostname and link_hostname.startswith(
                    ("ddownload", "rapidgator", "turbobit", "filecrypt")
                ):
                    if "rapidgator" in link_hostname:
                        links.insert(0, [resolved_url, link_hostname])
                    else:
                        links.append([resolved_url, link_hostname])

        except Exception as e:
            info(f"Error loading download links: {e}")
            mark_hostname_issue(
                Source.initials,
                "download",
                str(e) if "e" in dir() else "Download error",
            )

        return {"links": links}



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re

from quasarr.providers.imdb_metadata import get_imdb_id_from_title, get_poster_link


def canonicalize_solver_name(name):
    """Normalise a solver/provider name and return (key, display_name)."""
    raw = str(name).strip() if name is not None else ""
    compact = raw.lower().replace(" ", "").replace("-", "").replace("_", "")
    if compact in {"2captcha", "2captchacom"}:
        return "2captcha", "2Captcha"
    if compact in {"dbc", "deathbycaptcha", "deathbycaptchacom"}:
        return "deathbycaptcha", "DeathByCaptcha"
    if not raw:
        return "unknown", "Unknown"
    return compact, raw


def format_duration(value):
    """Format a duration value (seconds or HH:MM:SS string) into HH:MM:SS."""
    if value is None:
        return None
    if isinstance(value, str):
        stripped = value.strip()
        if not stripped:
            return None
        if stripped.count(":") == 2:
            try:
                parts = [int(part) for part in stripped.split(":")]
                if len(parts) == 3:
                    hours, minutes, seconds = parts
                    total_seconds = max(0, (hours * 3600) + (minutes * 60) + seconds)
                    hours = total_seconds // 3600
                    minutes = (total_seconds % 3600) // 60
                    seconds = total_seconds % 60
                    return f"{hours:02d}:{minutes:02d}:{seconds:02d}"
            except ValueError:
                return None
        match = re.fullmatch(r"(\d+(?:\.\d+)?)\s*[sS]?", stripped)
        if match:
            value = float(match.group(1))
        else:
            return None
    if isinstance(value, (int, float)):
        total_seconds = max(0, int(round(float(value))))
        hours = total_seconds // 3600
        minutes = (total_seconds % 3600) // 60
        seconds = total_seconds % 60
        return f"{hours:02d}:{minutes:02d}:{seconds:02d}"
    return None


def format_number(value):
    """Format a numeric value to string, stripping trailing zeros for floats."""
    if value is None:
        return None
    if isinstance(value, (int, float)):
        if isinstance(value, float):
            return f"{value:.5f}".rstrip("0").rstrip(".")
        return str(value)
    return str(value)


def format_balance(value):
    """Format a balance value to two decimal places."""
    if value is None:
        return None
    if isinstance(value, (int, float)):
        return f"{value:.2f}"
    try:
        return f"{float(value):.2f}"
    except (TypeError, ValueError):
        return str(value)


def build_solved_data(details):
    """Extract and group solver data from a solved-CAPTCHA details dict.

    Returns a list of dicts with keys: solver_display, attempts, cost, balance,
    currency, duration.  Returns None when no meaningful data is present.
    """
    if not isinstance(details, dict):
        return None

    providers = details.get("solvers")
    grouped_providers = []
    grouped_provider_map = {}
    if isinstance(providers, list):
        for provider in providers:
            if not isinstance(provider, dict):
                continue
            raw_name = provider.get("name")
            solver_key, solver_display = canonicalize_solver_name(raw_name)
            if solver_key not in grouped_provider_map:
                grouped_provider_map[solver_key] = {
                    "solver_key": solver_key,
                    "solver_display": solver_display,
                    "attempts": 0,
                    "has_attempts": False,
                    "cost": None,
                    "balance": None,
                    "currency": None,
                }
                grouped_providers.append(grouped_provider_map[solver_key])

            grouped_provider = grouped_provider_map[solver_key]

            if provider.get("attempts") is not None:
                try:
                    grouped_provider["attempts"] += max(
                        0, int(provider.get("attempts"))
                    )
                    grouped_provider["has_attempts"] = True
                except (TypeError, ValueError):
                    pass
            if provider.get("cost") is not None:
                grouped_provider["cost"] = provider.get("cost")
            if provider.get("balance") is not None:
                grouped_provider["balance"] = provider.get("balance")
            if provider.get("currency"):
                grouped_provider["currency"] = provider.get("currency")

    if not grouped_providers:
        result = {}
    else:
        preferred_order = {"2captcha": 0, "deathbycaptcha": 1}
        sorted_providers = sorted(
            grouped_providers,
            key=lambda p: (
                preferred_order.get(p.get("solver_key"), 99),
                p.get("solver_key", ""),
            ),
        )
        result = {"solvers": sorted_providers}

    duration_value = details.get("duration_seconds")
    formatted_duration = format_duration(duration_value)
    if formatted_duration:
        result["duration"] = formatted_duration

    return result or None


def resolve_poster_url(shared_state, title, imdb_id):
    """Resolve a poster image URL for the given title/IMDb ID.

    Returns the poster URL string or None.
    """
    if not imdb_id and " " not in title:
        imdb_id = get_imdb_id_from_title(shared_state, title)
    if imdb_id:
        poster_link = get_poster_link(shared_state, imdb_id)
        if poster_link:
            return poster_link
    return None



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import json
from html import escape
from urllib.parse import urlparse

import requests

from quasarr.constants import SESSION_REQUEST_TIMEOUT_SECONDS
from quasarr.providers.log import info
from quasarr.providers.notifications.helpers.abstract_notification_formatter import (
    AbstractNotificationFormatter,
)
from quasarr.providers.notifications.helpers.notification_message import (
    NotificationFactsEntry,
    NotificationLinkEntry,
    NotificationMessage,
    NotificationTextEntry,
    NotificationValueEntry,
)


def _get_telegram_credentials(shared_state):
    settings = shared_state.values.get("notification_settings")
    if not isinstance(settings, dict):
        return "", ""
    bot_token = str(settings.get("telegram_bot_token") or "").strip()
    chat_id = str(settings.get("telegram_chat_id") or "").strip()
    return bot_token, chat_id


def _escape_html_text(value):
    return escape(str(value), quote=False)


def _escape_html_attribute(value):
    return escape(str(value), quote=True)


def _get_photo_request_headers(shared_state, image_url):
    headers = {"Accept": "image/*,*/*;q=0.8"}
    user_agent = shared_state.values.get("user_agent")
    if user_agent:
        headers["User-Agent"] = user_agent

    hostname = (urlparse(image_url).hostname or "").lower()
    if hostname.endswith("media-amazon.com") or hostname.endswith("media-imdb.com"):
        headers["Referer"] = "https://www.imdb.com/"

    return headers


def _build_photo_upload(shared_state, image_url):
    response = requests.get(
        image_url,
        headers=_get_photo_request_headers(shared_state, image_url),
        timeout=SESSION_REQUEST_TIMEOUT_SECONDS,
    )
    response.raise_for_status()

    content_type = response.headers.get("Content-Type", "").split(";", 1)[0].strip()
    if not content_type.startswith("image/"):
        raise ValueError(f"Unexpected poster content type: {content_type or 'unknown'}")

    image_bytes = response.content
    if not image_bytes:
        raise ValueError("Poster download returned an empty response")

    parsed_url = urlparse(image_url)
    filename = parsed_url.path.rsplit("/", 1)[-1] or "poster.jpg"
    if "." not in filename:
        filename += ".jpg"

    return {"photo": (filename, image_bytes, content_type)}


def _parse_json_response(response):
    try:
        return response.json()
    except ValueError:
        return {}


def _telegram_get(api_base_url, method, params):
    response = requests.get(
        f"{api_base_url}/{method}",
        params=params,
        timeout=SESSION_REQUEST_TIMEOUT_SECONDS,
    )
    return response, _parse_json_response(response)


def _telegram_post(api_base_url, method, payload, files=None):
    response = requests.post(
        f"{api_base_url}/{method}",
        data=payload,
        files=files,
        timeout=SESSION_REQUEST_TIMEOUT_SECONDS,
    )
    return response, _parse_json_response(response)


def _extract_chat_capabilities(api_base_url, chat_id):
    chat_response, chat_result = _telegram_get(
        api_base_url, "getChat", {"chat_id": chat_id}
    )
    if not chat_result.get("ok"):
        description = chat_result.get("description", chat_response.status_code)
        return {
            "ok": False,
            "message": f"Telegram chat inspection failed: {description}",
        }

    chat = chat_result.get("result") or {}
    permissions = chat.get("permissions") or {}
    chat_type = str(chat.get("type") or "unknown")

    me_response, me_result = _telegram_get(api_base_url, "getMe", {})
    if not me_result.get("ok"):
        description = me_result.get("description", me_response.status_code)
        return {
            "ok": False,
            "message": f"Telegram bot inspection failed: {description}",
            "chat_type": chat_type,
        }

    bot_id = (me_result.get("result") or {}).get("id")
    member = {}
    status = None
    if bot_id is not None:
        member_response, member_result = _telegram_get(
            api_base_url,
            "getChatMember",
            {"chat_id": chat_id, "user_id": bot_id},
        )
        if member_result.get("ok"):
            member = member_result.get("result") or {}
            status = member.get("status")
        else:
            status = "unknown"
            description = member_result.get("description", member_response.status_code)
            return {
                "ok": False,
                "message": f"Telegram chat member inspection failed: {description}",
                "chat_type": chat_type,
            }

    can_send_photos = None
    can_add_previews = None

    if chat_type == "private":
        can_send_photos = True
        can_add_previews = True
    elif status in {"administrator", "creator"}:
        can_send_photos = True
        can_add_previews = True
    elif status == "restricted":
        can_send_photos = bool(member.get("can_send_photos", False))
        can_add_previews = bool(member.get("can_add_web_page_previews", False))
    elif status in {"left", "kicked"}:
        can_send_photos = False
        can_add_previews = False
    else:
        if "can_send_photos" in permissions:
            can_send_photos = bool(permissions.get("can_send_photos"))
        if "can_add_web_page_previews" in permissions:
            can_add_previews = bool(permissions.get("can_add_web_page_previews"))

    return {
        "ok": True,
        "chat_type": chat_type,
        "member_status": status,
        "can_send_photos": can_send_photos,
        "can_add_web_page_previews": can_add_previews,
    }


def _describe_chat_capabilities(capabilities):
    if not isinstance(capabilities, dict):
        return "Telegram destination capabilities are unknown."

    if not capabilities.get("ok"):
        return capabilities.get("message") or "Telegram destination inspection failed."

    chat_type = capabilities.get("chat_type") or "chat"
    member_status = capabilities.get("member_status") or "unknown"
    can_send_photos = capabilities.get("can_send_photos")
    can_add_previews = capabilities.get("can_add_web_page_previews")

    if can_send_photos is False and can_add_previews is False:
        return (
            f"Telegram {chat_type} permissions block both photos and web page previews "
            f"for bot status '{member_status}'. Enable media permissions or promote the bot."
        )
    if can_send_photos is False and can_add_previews is True:
        return (
            f"Telegram {chat_type} permissions block photos for bot status "
            f"'{member_status}'. Quasarr will use a large link preview for the poster."
        )
    if can_send_photos is True:
        return (
            f"Telegram {chat_type} permissions allow photo delivery for bot status "
            f"'{member_status}'."
        )
    if can_add_previews is True:
        return (
            f"Telegram {chat_type} permissions allow link previews for bot status "
            f"'{member_status}'."
        )
    return (
        f"Telegram {chat_type} permissions could not be determined for bot status "
        f"'{member_status}'."
    )


def _allow_link_preview(capabilities):
    return capabilities.get("can_add_web_page_previews") is not False


def _send_text_message(
    api_base_url,
    chat_id,
    text,
    disable_notification,
    image_url=None,
    use_link_preview=False,
):
    message_payload = {
        "chat_id": chat_id,
        "text": text,
        "parse_mode": "HTML",
        "disable_notification": disable_notification,
    }
    if image_url and use_link_preview:
        message_payload["link_preview_options"] = json.dumps(
            {
                "url": image_url,
                "prefer_large_media": True,
                "show_above_text": True,
            }
        )
    else:
        message_payload["disable_web_page_preview"] = True

    response, result = _telegram_post(api_base_url, "sendMessage", message_payload)
    if not result.get("ok"):
        description = result.get("description", response.status_code)
        info(f"Failed to send Telegram notification: {description}")
        return False
    return True


def inspect_destination(shared_state):
    bot_token, chat_id = _get_telegram_credentials(shared_state)
    if not bot_token or not chat_id:
        return {
            "ok": False,
            "message": "Telegram bot token and chat ID are required.",
        }

    api_base_url = f"https://api.telegram.org/bot{bot_token}"
    capabilities = _extract_chat_capabilities(api_base_url, chat_id)
    capabilities["message"] = _describe_chat_capabilities(capabilities)
    return capabilities


class TelegramNotificationFormatter(AbstractNotificationFormatter):
    @staticmethod
    def _format_link(text, url, link_text=None):
        safe_url = _escape_html_attribute(url)
        target_text = link_text or text
        safe_target_text = _escape_html_text(target_text)
        rendered_link = f'<a href="{safe_url}">{safe_target_text}</a>'
        if target_text and target_text in text:
            return _escape_html_text(text).replace(safe_target_text, rendered_link, 1)
        return rendered_link

    @staticmethod
    def _render_titled_entry(title, value):
        return f"<b>{_escape_html_text(title)}</b>\n{value}"

    def render_text_entry(self, entry: NotificationTextEntry):
        return self._render_titled_entry(entry.title, _escape_html_text(entry.text))

    def render_link_entry(self, entry: NotificationLinkEntry):
        return self._render_titled_entry(
            entry.title,
            self._format_link(entry.text, entry.url, entry.link_text),
        )

    def render_facts_entry(self, entry: NotificationFactsEntry):
        facts_text = " | ".join(
            f"<b>{_escape_html_text(fact.label)}:</b> {_escape_html_text(fact.value)}"
            for fact in entry.facts
        )
        return self._render_titled_entry(entry.title, facts_text)

    def render_value_entry(self, entry: NotificationValueEntry):
        return self._render_titled_entry(entry.title, _escape_html_text(entry.value))

    def render_message(self, message: NotificationMessage):
        parts = [
            f"<b>{_escape_html_text(message.title)}</b>",
            _escape_html_text(message.description),
        ]
        parts.extend(self.render_entries(message.entries))
        return "\n\n".join(parts)


def send(shared_state, message, silent=True):
    """Send a rendered Telegram notification. Returns True on success."""
    bot_token, chat_id = _get_telegram_credentials(shared_state)
    if not bot_token or not chat_id:
        return False

    if not isinstance(message, NotificationMessage):
        info(f"Invalid Telegram notification payload: {type(message).__name__}")
        return False

    text = TelegramNotificationFormatter().render_message(message)
    disable_notification = bool(silent)
    api_base_url = f"https://api.telegram.org/bot{bot_token}"
    capabilities = inspect_destination(shared_state)

    if message.image_url:
        photo_allowed = capabilities.get("can_send_photos")
        preview_allowed = _allow_link_preview(capabilities)

        if photo_allowed is False:
            info(capabilities.get("message") or "Telegram chat blocks photo delivery.")
            return _send_text_message(
                api_base_url,
                chat_id,
                text,
                disable_notification,
                image_url=message.image_url,
                use_link_preview=preview_allowed,
            )

        photo_payload = {
            "chat_id": chat_id,
            "disable_notification": disable_notification,
        }
        if len(text) <= 1024:
            photo_payload["caption"] = text
            photo_payload["parse_mode"] = "HTML"

        try:
            photo_files = _build_photo_upload(shared_state, message.image_url)
        except Exception as e:
            info(f"Failed to fetch Telegram photo notification image: {e}")
            return _send_text_message(
                api_base_url,
                chat_id,
                text,
                disable_notification,
                image_url=message.image_url,
                use_link_preview=preview_allowed,
            )

        photo_response, photo_result = _telegram_post(
            api_base_url, "sendPhoto", photo_payload, files=photo_files
        )
        if not photo_result.get("ok"):
            description = photo_result.get("description", photo_response.status_code)
            info(f"Failed to send Telegram photo notification: {description}")
            if preview_allowed:
                return _send_text_message(
                    api_base_url,
                    chat_id,
                    text,
                    disable_notification,
                    image_url=message.image_url,
                    use_link_preview=True,
                )
            return _send_text_message(
                api_base_url,
                chat_id,
                text,
                disable_notification,
            )
        if len(text) <= 1024:
            return True

    return _send_text_message(api_base_url, chat_id, text, disable_notification)



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

import re
from datetime import datetime
from urllib.parse import unquote, urljoin, urlparse

import requests

from quasarr.constants import (
    DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
    SHARE_HOSTERS_LOWERCASE,
)
from quasarr.downloads.sources.helpers.abstract_source import AbstractDownloadSource
from quasarr.providers.hostname_issues import clear_hostname_issue, mark_hostname_issue
from quasarr.providers.log import info

_IMDB_ID_REGEX = re.compile(r"(tt\d+)", re.I)


class Source(AbstractDownloadSource):
    initials = "rm"

    def get_download_links(self, shared_state, url, mirrors, title, password):
        slug = _extract_release_slug(url)
        if not slug:
            mark_hostname_issue(Source.initials, "download", "Invalid release URL")
            return {"links": [], "imdb_id": None}

        try:
            payload = _fetch_release_payload(
                shared_state,
                slug,
                DOWNLOAD_REQUEST_TIMEOUT_SECONDS,
            )
            release = payload.get("release") or {}
            links = _extract_download_links(release.get("links") or [], mirrors)
            imdb_id = _extract_imdb_id(payload.get("production"))

            if links:
                clear_hostname_issue(Source.initials)
            else:
                mark_hostname_issue(
                    Source.initials,
                    "download",
                    "No supported download links found",
                )

            return {
                "links": links,
                "imdb_id": imdb_id,
                "title": release.get("title"),
            }
        except Exception as e:
            info(f"Could not load release data for {title or slug}: {e}")
            mark_hostname_issue(
                Source.initials,
                "download",
                str(e) if "e" in dir() else "Download error",
            )
            return {"links": [], "imdb_id": None}


def _get_base_url(shared_state):
    host = shared_state.values["config"]("Hostnames").get(Source.initials)
    if not host:
        return None

    normalized = str(host).strip()
    if "://" in normalized:
        parsed = urlparse(normalized)
        normalized = parsed.netloc or parsed.path

    normalized = normalized.rstrip("/")
    if normalized.startswith("www."):
        normalized = normalized[4:]
    if not normalized.startswith("v2."):
        normalized = f"v2.{normalized}"

    return f"https://{normalized}"


def _build_headers(shared_state):
    return {
        "Accept": "application/json",
        "User-Agent": shared_state.values["user_agent"],
    }


def _build_release_page_url(base_url, slug):
    return urljoin(f"{base_url}/", f"release/{slug}")


def _convert_iso_to_rss_date(iso_date):
    if not iso_date:
        return ""

    parsed_date = datetime.fromisoformat(str(iso_date).replace("Z", "+00:00"))
    return parsed_date.strftime("%a, %d %b %Y %H:%M:%S %z")


def _is_movie_release(release):
    return release.get("season") is None and release.get("episode") is None


def _fetch_release_payload(shared_state, slug, timeout):
    base_url = _get_base_url(shared_state)
    if not base_url:
        raise ValueError("Missing hostname configuration")

    response = requests.get(
        f"{base_url}/api/releases/by-slug/{slug}",
        headers=_build_headers(shared_state),
        timeout=timeout,
    )
    response.raise_for_status()
    return response.json()


def _create_session(shared_state):
    session = requests.Session()
    session.headers.update({"User-Agent": shared_state.values["user_agent"]})
    return session


def _bootstrap_session(shared_state, timeout):
    base_url = _get_base_url(shared_state)
    if not base_url:
        raise ValueError("Missing hostname configuration")

    session = _create_session(shared_state)
    response = session.get(
        f"{base_url}/",
        timeout=timeout,
    )
    response.raise_for_status()
    return session


def _get_xsrf_token(session):
    return unquote(session.cookies.get("XSRF-TOKEN", ""))


def _fetch_releases_by_production_ids(
    shared_state,
    production_ids,
    timeout,
    per_page=100,
    session=None,
):
    base_url = _get_base_url(shared_state)
    if not base_url:
        raise ValueError("Missing hostname configuration")

    normalized_ids = [str(pid).strip() for pid in production_ids if str(pid).strip()]
    if not normalized_ids:
        return []

    session = session or _bootstrap_session(shared_state, timeout)
    response = session.post(
        f"{base_url}/api/releases/by-production-ids",
        params={
            "page": 1,
            "per_page": per_page,
            "sort_field": "updated_at",
            "sort_direction": "desc",
        },
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
            "X-XSRF-TOKEN": _get_xsrf_token(session),
        },
        json={"production_ids": ",".join(normalized_ids)},
        timeout=timeout,
    )
    response.raise_for_status()
    return (response.json().get("releases") or {}).get("data") or []


def _search_productions(shared_state, query, timeout, limit=10, session=None):
    base_url = _get_base_url(shared_state)
    if not base_url:
        raise ValueError("Missing hostname configuration")

    requester = session or requests
    response = requester.get(
        f"{base_url}/api/productions/search",
        headers=_build_headers(shared_state),
        params={"q": str(query or "").strip(), "limit": limit},
        timeout=timeout,
    )
    response.raise_for_status()
    payload = response.json()
    return (payload.get("results") or {}).get("productions") or []


def _fetch_productions_page(shared_state, page, timeout, per_page=20, session=None):
    base_url = _get_base_url(shared_state)
    if not base_url:
        raise ValueError("Missing hostname configuration")

    requester = session or requests
    response = requester.get(
        f"{base_url}/api/productions/productions",
        headers=_build_headers(shared_state),
        params={
            "sort_by": "rmz_updated_at",
            "sort_direction": "desc",
            "per_page": per_page,
            "page": page,
        },
        timeout=timeout,
    )
    response.raise_for_status()
    return response.json()


def _extract_release_slug(url):
    try:
        parsed = urlparse(str(url or "").strip())
    except Exception:
        return None

    if not parsed.path:
        return None

    path = parsed.path.rstrip("/")
    if "/release/" in path:
        return unquote(path.rsplit("/", 1)[-1])

    marker = "/api/releases/by-slug/"
    if marker in path:
        return unquote(path.split(marker, 1)[1].split("/", 1)[0])

    return unquote(path.rsplit("/", 1)[-1]) if path else None


def _normalize_requested_mirror(mirror_name):
    normalized = str(mirror_name or "").lower().strip()

    if "://" in normalized:
        parsed = urlparse(normalized)
        normalized = parsed.netloc or parsed.path

    normalized = normalized.split("/", 1)[0]
    normalized = normalized.split(":", 1)[0]
    if normalized.startswith("www."):
        normalized = normalized[4:]
    if "." in normalized:
        normalized = normalized.split(".", 1)[0]

    return normalized


def _derive_supported_mirror(link):
    host = link.get("host") or {}
    provider = str(host.get("slug") or "").lower().strip()

    if not provider:
        try:
            hostname = (urlparse(link.get("url") or "").hostname or "").lower()
        except Exception:
            hostname = ""
        if hostname.startswith("www."):
            hostname = hostname[4:]
        provider = hostname.split(".", 1)[0]

    if provider in SHARE_HOSTERS_LOWERCASE:
        return provider
    return None


def _extract_download_links(release_links, mirrors=None):
    requested_mirrors = {
        _normalize_requested_mirror(mirror) for mirror in (mirrors or []) if mirror
    }

    links = []
    seen = set()

    for link in release_links:
        href = str(link.get("url") or "").strip()
        if not href:
            continue

        mirror = _derive_supported_mirror(link)
        if not mirror:
            continue

        if requested_mirrors and mirror not in requested_mirrors:
            continue

        if href in seen:
            continue
        seen.add(href)
        links.append([href, mirror])

    return links


def _extract_imdb_id(production):
    if not production:
        return None

    for website in production.get("websites") or []:
        match = _IMDB_ID_REGEX.search(str(website.get("url") or ""))
        if match:
            return match.group(1)

    return None



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from abc import ABC
from dataclasses import dataclass


@dataclass(frozen=True)
class NotificationFact:
    label: str
    value: str


@dataclass(frozen=True)
class AbstractNotificationEntry(ABC):
    title: str


@dataclass(frozen=True)
class NotificationTextEntry(AbstractNotificationEntry):
    text: str


@dataclass(frozen=True)
class NotificationLinkEntry(AbstractNotificationEntry):
    text: str
    url: str
    link_text: str | None = None


@dataclass(frozen=True)
class NotificationFactsEntry(AbstractNotificationEntry):
    facts: tuple[NotificationFact, ...]


@dataclass(frozen=True)
class NotificationValueEntry(AbstractNotificationEntry):
    value: str


@dataclass(frozen=True)
class NotificationMessage:
    title: str
    description: str
    entries: tuple[AbstractNotificationEntry, ...] = ()
    image_url: str | None = None
    thumbnail_url: str | None = None



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from bottle import request, response

from quasarr.constants import (
    TIMEOUT_SLOW_MODE_DEFINITIONS,
    TIMEOUT_SLOW_MODE_TABLE,
    apply_timeout_slow_mode_settings,
)
from quasarr.storage.sqlite_database import DataBase


def _coerce_bool(value, default=False):
    if isinstance(value, bool):
        return value
    if value is None:
        return default
    if isinstance(value, str):
        lowered = value.strip().lower()
        if lowered in {"1", "true", "yes", "on"}:
            return True
        if lowered in {"0", "false", "no", "off"}:
            return False
    return bool(value)


def _read_timeout_slow_mode_settings():
    timeout_settings_db = DataBase(TIMEOUT_SLOW_MODE_TABLE)

    settings = {}
    for timeout_key in TIMEOUT_SLOW_MODE_DEFINITIONS:
        settings[timeout_key] = _coerce_bool(
            timeout_settings_db.retrieve(timeout_key),
            default=False,
        )

    return settings


def refresh_timeout_slow_mode_settings(shared_state):
    settings = _read_timeout_slow_mode_settings()
    shared_state.update("timeout_slow_mode", settings)
    apply_timeout_slow_mode_settings(settings)
    return settings


def initialize_timeout_slow_mode_settings(shared_state):
    return refresh_timeout_slow_mode_settings(shared_state)


def get_timeout_slow_mode_settings_data(shared_state):
    response.content_type = "application/json"
    settings = refresh_timeout_slow_mode_settings(shared_state)
    return {"success": True, "settings": settings}


def save_timeout_slow_mode_settings(shared_state):
    response.content_type = "application/json"

    data = request.json
    if not isinstance(data, dict):
        return {"success": False, "message": "Invalid JSON payload"}

    provided_settings = data.get("settings")
    if not isinstance(provided_settings, dict):
        provided_settings = {}

    timeout_settings_db = DataBase(TIMEOUT_SLOW_MODE_TABLE)
    current_settings = _read_timeout_slow_mode_settings()

    for timeout_key in TIMEOUT_SLOW_MODE_DEFINITIONS:
        current_value = current_settings.get(timeout_key, False)
        next_value = _coerce_bool(provided_settings.get(timeout_key), current_value)
        timeout_settings_db.update_store(
            timeout_key,
            "true" if next_value else "false",
        )

    settings = refresh_timeout_slow_mode_settings(shared_state)

    return {
        "success": True,
        "message": "Timeout slow mode settings saved successfully",
        "settings": settings,
    }



# -*- coding: utf-8 -*-
# Quasarr
# Project by https://github.com/rix1337

from enum import Enum


class NotificationType(str, Enum):
    UNPROTECTED = "unprotected"
    CAPTCHA = "captcha"
    SOLVED = "solved"
    DISABLED = "disabled"
    FAILED = "failed"
    QUASARR_UPDATE = "quasarr_update"
    TEST = "test"


_NOTIFICATION_TYPE_LABELS = {
    NotificationType.UNPROTECTED: "Unprotected Links",
    NotificationType.CAPTCHA: "CAPTCHA Required",
    NotificationType.SOLVED: "CAPTCHA Solved",
    NotificationType.DISABLED: "SponsorsHelper Disabled",
    NotificationType.FAILED: "SponsorsHelper Failed",
    NotificationType.QUASARR_UPDATE: "Quasarr Updates",
    NotificationType.TEST: "Test Message",
}


def normalize_notification_type(value):
    if isinstance(value, NotificationType):
        return value

    if value is None:
        return None

    normalized_value = str(value).strip().lower()
    for notification_type in NotificationType:
        if notification_type.value == normalized_value:
            return notification_type

    return None


def get_user_configurable_notification_types():
    return tuple(
        notification_type
        for notification_type in NotificationType
        if notification_type is not NotificationType.TEST
    )


def get_notification_type_label(notification_type):
    return _NOTIFICATION_TYPE_LABELS.get(
        notification_type,
        notification_type.value.replace("_", " ").title(),
    )
